# -*- coding: UTF-8 -*-
# @Time : 2018/4/28
# @File : queue_dataset.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.dataflow.dataset.dataset import Dataset
import os
import sys
from antgo.utils.serialize import *
from io import BytesIO
import numpy as np
import cv2

class QueueDataset(Dataset):
  def __init__(self, queue=None):
    super(QueueDataset, self).__init__('test', '', None)
    self.queue = queue
    self.annotation = {}

  @property
  def size(self):
    return sys.maxsize

  def at(self, id):
    raise NotImplementedError

  def now(self):
    return self.annotation

  def _parse_data(self, data, data_type, request_param):
    assert(data_type not in ['VIDEO'])
    try:
      if data_type == 'IMAGE':
        img_data = cv2.imread(data)
        return (img_data,{}) if request_param is None else (img_data, request_param)
      if data_type == 'IMAGE_MEMORY':
        img_data = cv2.imread(BytesIO(data))
        return (img_data,{}) if request_param is None else (img_data, request_param)
      elif data_type == 'FILE':
        with open(data, 'r') as fp:
          return (fp.read(),{}) if request_param is None else (fp.read(), request_param)
      elif data_type == 'STRING':
        return (data,{}) if request_param is None else (data, request_param)
      elif data_type == 'JSON':
        return (data,{}) if request_param is None else (data, request_param)
      else:
        return (None,{})
    except:
      return None

  def data_pool(self):
    while True:
      data_pack = self.queue.get()
      if type(data_pack) == list and \
              (type(data_pack[0]) == list or type(data_pack[0]) == tuple):
        # multi-data
        for single_pack in data_pack:
          data, data_type, request_param = single_pack
          response_data = self._parse_data(data, data_type, request_param)
          if response_data is None:
            # 解析数据有误，不进行处理
            continue

          # annotation
          self.annotation = response_data[1]
          yield response_data
      else:
        # single-data
        data, data_type, request_param = data_pack
        if data_type == 'VIDEO':
          # 解析视频文件
          try:
            cap = cv2.VideoCapture(data)

            if request_param is None:
              request_param = {}

            # 视频fps
            fps = cap.get(5)
            request_param.update({'fps': fps})

            # 视频帧数
            frame_num=cap.get(7)
            request_param.update({'frame_num': (int)(frame_num)})

            frame_index = 0       
            while True:
              ret, frame = cap.read()
              if not ret:
                break
              
              request_param.update({'frame_index': frame_index})
              self.annotation = request_param
              yield (frame, {}) if request_param is None else (frame, request_param)
              frame_index += 1              
              if frame_index == (int)(frame_num):
                break
          except:
            # 解析视频有误，不进行处理
            continue
        else:
          response_data = self._parse_data(data, data_type, request_param)
          if response_data is None:
            # 解析数据有误，不进行处理
            continue

          # annotation
          self.annotation = response_data[1]
          yield response_data


# encoding=utf-8
# @Time     : 17-8-14
# @File     : serialize.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import importlib
import msgpack
import msgpack_numpy
msgpack_numpy.patch()

import json
__all__ = ['loads', 'dumps', 'Encoder', 'Decoder']


def dumps(obj):
  return msgpack.dumps(obj)


def loads(buf):
  return msgpack.loads(buf)


class Decoder(json.JSONDecoder):
  def __init__(self):
    json.JSONDecoder.__init__(self, object_hook=self.dict_to_object)

  def dict_to_object(self, d):
    if '__class__' in d:
      class_name = d.pop('__class__')
      module_name = d.pop('__module__')
      # module = __import__(module_name.split('.')[1])
      # module = __import__(module_name)
      module = importlib.import_module(module_name)
      class_ = getattr(module, class_name)
      args = dict((key, value) for key, value in d.items())

      inst = class_(**args)
    else:
      inst = d
    return inst


class Encoder(json.JSONEncoder):
  def default(self, obj):
    # Convert objects to a dictionary of their representation
    d = {'__class__': obj.__class__.__name__,
         '__module__': obj.__module__,
         }
    d.update(obj.__dict__)
    return d



# -*- coding: UTF-8 -*-
# @Time : 2018/8/13
# @File : ade20k.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset import *
import os
import numpy as np

__all__ = ['ADE20K']
class ADE20K(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(ADE20K, self).__init__(train_or_test, dir)

    assert(train_or_test in ['sample', 'train', 'val', 'test'])
    if train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    self._image_file_list = []
    self._annotation_file_list = []
    if train_or_test in ['train', 'val']:
      image_target_path = os.path.join(dir,
                                       'ADEChallengeData2016',
                                       'images',
                                       'training' if train_or_test == 'train' else 'validation')
      annotation_target_path = os.path.join(dir,
                                            'ADEChallengeData2016',
                                            'annotations',
                                            'training' if train_or_test == 'train' else 'validation')
      if not os.path.exists(image_target_path):
        logger.error('ADE dtaset must download from \n http://data.csail.mit.edu/places/ADEchallenge/ADEChallengeData2016.zip')
        sys.exit(0)

      for file in os.listdir(image_target_path):
        if file[0] == '.':
          continue

        self._image_file_list.append(os.path.join(image_target_path, file))
        self._annotation_file_list.append(os.path.join(annotation_target_path, '%s.png'%file.split('.')[0]))
    else:
      parse_file = os.path.join(dir, 'release_test', 'list.txt')
      if not os.path.exists(parse_file):
        logger.error(
          'ADE dtaset must download from \n http://data.csail.mit.edu/places/ADEchallenge/release_test.zip')
        sys.exit(0)

      with open(parse_file) as fp:
        for file in fp.readlines():
          if len(file) == 0:
            continue

          self._image_file_list.append(os.path.join(dir, 'release_test', 'testing', file.replace('\n','')))

    self.ids = list(range(len(self._image_file_list)))

  @property
  def size(self):
    return len(self.ids)

  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.copy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]

      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(idxs)

      for k in idxs:
        image_file = self._image_file_list[k]
        image = imread(image_file)

        if self.train_or_test in ['train', 'val']:
          label_file = self._annotation_file_list[k]
          label = imread(label_file)

          if len(label.shape) == 3:
            label = label[:,:,0]

          yield [image, {'segmentation_map': label, 'file_id': image_file.split('/')[-1]}]
        else:
          yield [image, {'file_id': image_file.split('/')[-1]}]

  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    image_file = self._image_file_list[id]
    image = imread(image_file)

    if self.train_or_test in ['train', 'val']:
      label_file = self._annotation_file_list[id]
      label = imread(label_file)

      if len(label.shape) == 3:
        label = label[:, :, 0]

      return [image, {'segmentation_map':label, 'file_id': image_file.split('/')[-1]}]
    else:
      return [image, {'file_id': image_file.split('/')[-1]}]

  def split(self, split_params={}, split_method='holdout'):
    assert(self.train_or_test == 'train')
    assert(split_method == 'holdout')

    validation_dataet = ADE20K('val', self.dir)
    return self, validation_dataet



# -*- coding: UTF-8 -*-
# @Time    : 18-5-3
# @File    : empty_dataset.py
# @Author  : 
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset.dataset import Dataset
import os
import copy
import sys
import time


class EmptyDataset(Dataset):
  def __init__(self, train_or_test, dir, params={}):
    super(EmptyDataset, self).__init__(train_or_test, dir, params)

  @property
  def size(self):
    return 0

  def at(self, id):
    raise NotImplementedError

  def data_pool(self):
    time.sleep(sys.maxsize)
    yield None, None


# encoding=utf-8
# @Time    : 17-3-3
# @File    : main.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import sys
from typing import NamedTuple
# sys.path.append('/workspace/workspace/portrait_code/tool/antgo')
# sys.path.append('/workspace/workspace/portrait_code/tool/antgo/antgo')
from antgo.ant.generate import *
from antgo.ant.demo import *
from antgo.ant.train import *
from antgo.ant.challenge import *
from antgo.ant.browser import *
from antgo.ant.batch import *
from antgo.ant.activelearning import *
from antgo.ant import activelearning_api
from antgo.ant import spider_api
from antgo.ant.utils import *
from antgo.sandbox.sandbox import *
from antgo.utils.utils import *
from antgo.utils.dht import *
from antgo import version
import traceback
from jinja2 import Environment, FileSystemLoader

import multiprocessing
import subprocess

if sys.version > '3':
    PY3 = True
else:
    PY3 = False


def _check_environment():
  is_in_mltalker = True if os.environ.get('ANT_ENVIRONMENT', '') != '' else False
  return is_in_mltalker


_ant_support_commands = ["train",
                         "challenge",
                         "activelearning",
                         "dataset",
                         "predict",
                         "demo",
                         "spider",
                         "browser",
                         "startproject",
                         "config"]

#############################################
#######   antgo parameters            #######
#############################################
flags.DEFINE_string('main_file', None, 'main file')
flags.DEFINE_string('main_param', None, 'model parameters')
flags.DEFINE_string('main_folder', None, 'resource folder')
flags.DEFINE_string('version', None, 'minist antgo version')
flags.DEFINE_string('task', None, 'task file')
flags.DEFINE_string('task_t', '', 'task type')
flags.DEFINE_string('task_ep', '', 'holdout/repeated-holdout/bootstrap/kfold')
flags.DEFINE_string('task_em', '', 'define based on task type')
flags.DEFINE_integer('task_badcase_num', 100, 'badcase analysis (num)')
flags.DEFINE_integer('task_badcase_category', 1, 'badcase analysis (category)')
flags.DEFINE_string('dataset', None, 'dataset name')
flags.DEFINE_string('file', '', '')
flags.DEFINE_string('signature', '123', 'signature')
flags.DEFINE_string('devices', '', 'devices')
flags.DEFINE_string('servers', '', '')
flags.DEFINE_string('dump', None, 'dump dir')
flags.DEFINE_string('token', None, 'token')
flags.DEFINE_string('proxy', None, 'proxy')
flags.DEFINE_string('name', None, 'name')
flags.DEFINE_string('author', None, 'author')
flags.DEFINE_string('net', None, 'GENERAL')
flags.DEFINE_string('max_time', '100d', 'max running time')
flags.DEFINE_string('from_experiment', None, 'load model from experiment')
flags.DEFINE_string('restore_experiment', None, 'restore experiment')
flags.DEFINE_string('factory', None, '')
flags.DEFINE_string('config', None, 'config file')
flags.DEFINE_string('benchmark', None, 'benchmark experiments')
flags.DEFINE_string('host_ip', '127.0.0.1', 'host ip address')
flags.DEFINE_integer('host_port', -1, 'port')
flags.DEFINE_string('html_template', None, 'html template')
flags.DEFINE_indicator('worker', '')
flags.DEFINE_indicator('master', '')
flags.DEFINE_indicator('unlabel', '')           # 未标注数据
flags.DEFINE_indicator('candidate', '')         # 候选数据
flags.DEFINE_indicator('skip_training', '')
flags.DEFINE_indicator('research', 'research mode or not')
flags.DEFINE_string('running_platform', 'local', 'local/cloud')
flags.DEFINE_string('param', '', '')  # k:v;k:v


FLAGS = flags.AntFLAGS
Config = config.AntConfig


def main():
  # 1.step antgo logo
  main_logo()

  # 3.step parse antgo running params
  if len(sys.argv) >= 2:
    if sys.argv[1].startswith('--') or sys.argv[1].startswith('-'):
      flags.cli_param_flags(sys.argv[1:])
    else:
      flags.cli_param_flags(sys.argv[2:])

  # token
  token = FLAGS.token()
  if not PY3 and token is not None:
    token = unicode(token)

  if sys.argv[1] == 'config':
    if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo')):
      os.makedirs(os.path.join(os.environ['HOME'], '.config', 'antgo'))

    if FLAGS.config() is not None:
      # 使用指定配置文件更新
      shutil.copy(FLAGS.config(), os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml'))
      logger.info('Success update config file.')
    else:
      # 在当前目录生成默认配置文件
      # --dataset, --token, 自动配置
      config_data = {'DATASET_FACTORY': '', 'USER_TOKEN': ''}
      if FLAGS.dataset() is not None:
        if not os.path.exists(FLAGS.dataset()):
          logger.error('Dataset factory path dont exist.')
          return
        config_data['DATASET_FACTORY'] = FLAGS.dataset()

      if FLAGS.token() is not None:
         config_data['USER_TOKEN'] = FLAGS.token()

      env = Environment(loader=FileSystemLoader('/'.join(os.path.realpath(__file__).split('/')[0:-1])))
      config_template = env.get_template('config.xml')
      config_content = config_template.render(**config_data)

      if config_data['DATASET_FACTORY'] != '':
        with open(os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml'),'w') as fp:
          fp.write(config_content)
        
        logger.info('Finish antgo global config.')
      else:
        with open('./config.xml','w') as fp:
          fp.write(config_content)

        logger.info('Please fill ./config.xml, then call (antgo config --config=./config.xml) to finish global config.')
    return

  # 检查配置文件是否存在
  if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')):
    logger.error('Missing config file, please run antgo config.')
    return

  # 解析配置文件
  config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
  Config.parse_xml(config_xml)

  # 检查最低版本与当前版本的兼容性
  if FLAGS.running_platform == 'local' and FLAGS.version() is not None:
    a, b, c = FLAGS.version().split('.')
    sys_a, sys_b, sys_c = version.split('.')
    if int(sys_a) < int(a) or int(sys_b) < int(b) or int(sys_c) < int(c):
      logger.error('Antgo version dont satisfy task minimum request (%s).'%FLAGS.version())
      return

  # 4.step check factory
  factory = getattr(Config, 'factory', None)
  if factory is None or \
      factory == '' or not os.path.exists(Config.factory):
    logger.error('Factory folder is missing, please run antgo config.')
    return

  if not os.path.exists(Config.data_factory):
    os.makedirs(Config.data_factory)
  if not os.path.exists(Config.task_factory):
    os.makedirs(Config.task_factory)

  task_factory = Config.task_factory
  data_factory = Config.data_factory

  # 5.step parse antgo execute command
  # 5.1.step other command (train, challenge, compose, deploy, server)
  ant_cmd = sys.argv[1]
  ant_cmd = ant_cmd.split('/')[0]
  ant_cmd_api = '' if sys.argv[1] == ant_cmd else sys.argv[1].split('/')[-1]
  if ant_cmd not in _ant_support_commands:
    logger.error('Antgo cli support( %s )command.'%",".join(_ant_support_commands))
    return
  
  if ant_cmd_api != '':
    # api 
    if ant_cmd == 'activelearning':
      func = getattr(activelearning_api, 'activelearning_api_'+ant_cmd_api, None)
      if func is None:
        logger.error('%s/%s dont support.'%(ant_cmd, ant_cmd_api))
        return
      
      func()
    elif ant_cmd == 'spider':
      func = getattr(spider_api, 'spider_api_'+ant_cmd_api, None)
      if func is None:
        logger.error('%s/%s dont support.'%(ant_cmd, ant_cmd_api))
        return

      func()
    return

  if ant_cmd == 'startproject':
    project_name = FLAGS.name()
    if project_name == "":
      project_name = 'AntgoProject'

    if os.path.exists(os.path.join(os.curdir, project_name)):
      while True:
        project_i = 0
        if not os.path.exists(os.path.join(os.curdir, '%s-%d'%(project_name, project_i))):
          project_name = '%s-%d'%(project_name, project_i)
          break

        project_i += 1

    os.makedirs(os.path.join(os.curdir, project_name))

    # generate main file and main param templates
    template_file_folder = os.path.join(os.path.dirname(__file__), 'resource', 'templates')
    file_loader = FileSystemLoader(template_file_folder)
    env = Environment(loader=file_loader)
    template = env.get_template('task_main_file.template')
    output = template.render(ModelTime=datetime.fromtimestamp(timestamp()).strftime('%Y-%m-%d'),
                             ModelName=FLAGS.name(),
                             ModelAuthor=FLAGS.author() if FLAGS.author() is not None else 'xxx',
                             tensorflow=False,
                             GAN=False)

    with open(os.path.join(os.curdir, project_name, '%s_main.py' % FLAGS.name()),'w') as fp:
      fp.write(output)

    template = env.get_template('task_main_param.template')
    output = template.render(ModelTime=datetime.fromtimestamp(timestamp()).strftime('%Y-%m-%d'),
                             ModelName=FLAGS.name(),
                             ModelAuthor=FLAGS.author() if FLAGS.author() is not None else 'xxx')
    with open(os.path.join(os.curdir, project_name, '%s_param.yaml' % FLAGS.name()),'w') as fp:
      fp.write(output)

    template = env.get_template('task.template')
    output = template.render(ModelName=FLAGS.name())
    with open(os.path.join(os.curdir, project_name, '%s_task.xml' % FLAGS.name()),'w') as fp:
      fp.write(output)

    template = env.get_template('task_shell.template')
    output = ''
    if FLAGS.token() is not None:
      output = template.render(token=FLAGS.token(),
                               main_file='%s_main.py' % FLAGS.name(),
                               main_param='%s_param.yaml' % FLAGS.name())
    else:
      output = template.render(task='%s_task.xml' % FLAGS.name(),
                               main_file='%s_main.py' % FLAGS.name(),
                               main_param='%s_param.yaml' % FLAGS.name())

    with open(os.path.join(os.curdir, project_name, 'run.sh'), 'w') as fp:
      fp.write(output)

    return

  # 7.step check related params
  # 7.1 must set name
  name = FLAGS.name()
  if name is None:
    logger.error('Must set experiemnt name.')
    return

  # 7.2 check main folder (all related model code, includes main_file and main_param)
  main_folder = FLAGS.main_folder()
  if main_folder is None:
    main_folder = os.path.abspath(os.curdir)

  if os.path.exists(os.path.join(main_folder, '%s.tar.gz'%name)):
    untar_shell = 'openssl enc -d -aes256 -in %s.tar.gz -k %s | tar xz -C %s' % (name, FLAGS.signature(), '.')
    subprocess.call(untar_shell, shell=True, cwd=main_folder)
    os.remove(os.path.join(main_folder, '%s.tar.gz' % name))

    main_folder = os.path.join(main_folder, name)
    os.chdir(main_folder)

  # 7.3 check dump dir (all running data is stored here)
  dump_dir = FLAGS.dump()
  if dump_dir is None:
    dump_dir = os.path.join(os.path.abspath(os.curdir), 'dump')
    if not os.path.exists(dump_dir):
      os.makedirs(dump_dir)

  # 7.4 check main file
  main_file = FLAGS.main_file()
  if main_file is None or not os.path.exists(os.path.join(main_folder, main_file)):
    if not (FLAGS.worker() or FLAGS.master()):
      logger.error('Main executing file dont exist.')
      return

  # 8 step ant running
  # 8.1 step what is task
  task = FLAGS.task()
  if task is not None:
    if os.path.exists(os.path.join(task_factory, task)):
      task = os.path.join(task_factory, task)
    else:
      task = os.path.join(main_folder, task)

  dataset = FLAGS.dataset()
  if task is None and dataset is not None:
    # build default task
    with open(os.path.join(task_factory, '%s.xml'%name), 'w') as fp:
      task_content = '<task><task_name>%s</task_name><task_type>%s</task_type><task_badcase><badcase_num>%d</badcase_num><badcase_category>%d</badcase_category></task_badcase>' \
                     '<input>' \
                     '<source_data><data_set_name>%s</data_set_name>' \
                     '</source_data>' \
                     '<estimation_procedure><type>%s</type></estimation_procedure>' \
                     '<evaluation_measures><evaluation_measure>%s</evaluation_measure></evaluation_measures>' \
                     '</input>' \
                     '</task>'%(name, FLAGS.task_t(), FLAGS.task_badcase_num(), FLAGS.task_badcase_category(), dataset, FLAGS.task_ep(), FLAGS.task_em())
      fp.write(task_content)
    task = os.path.join(task_factory, '%s.xml'%name)

  # 8.2 step load ant context
  ant_context = None
  if main_file is not None and main_file != '':
    try:
      ant_context = main_context(main_file, main_folder)
    except Exception as e:
      traceback.print_exc()
      print('Fail to load main_file %s.'%main_file)
      return
  else:
    ant_context = Context()
  
  # 8.3 step load model config
  main_param = FLAGS.main_param()
  if main_param is not None:
    main_config_path = os.path.join(main_folder, main_param)
    try:
      with open(main_config_path, 'r') as fp:
        # 用户配置参数
        params = yaml.load(fp, Loader=yaml.FullLoader)

        # 系统全局参数
        params.update({
          'system': {'research': FLAGS.research(), 
                    'skip_training': FLAGS.skip_training(),
                    'ip': FLAGS.host_ip(),
                    'port': FLAGS.host_port(),
                    'devices': FLAGS.devices(),
                    'running_platform': FLAGS.running_platform(),
                    'unlabel': FLAGS.unlabel(),
                    'candidate': FLAGS.candidate()},
        })
        ant_context.params = params
    except Exception as e:
      traceback.print_exc()
      print('Fail to load main_param %s.'%main_param)
      return

  ant_context.name = name
  ant_context.data_factory = data_factory
  ant_context.task_factory = task_factory

  # 8.4 step load experiment
  if FLAGS.from_experiment() is not None and \
          FLAGS.running_platform() == 'local' and \
          ant_cmd not in ['release']:
    experiment_path = ''
    if os.path.isdir(FLAGS.from_experiment()):
      # 绝对路径指定的实验
      experiment_path = FLAGS.from_experiment()
    else:
      experiment_path = os.path.join(dump_dir, FLAGS.from_experiment())

    if not os.path.exists(experiment_path):
      logger.error('Couldnt find experiment %s.'%FLAGS.from_experiment())
      return

    # 设置实验目录
    ant_context.from_experiment = experiment_path

  # time stamp
  time_stamp = timestamp()
  if ant_cmd == "train":
    try:
      with running_sandbox(sandbox_time=FLAGS.max_time(),
                          sandbox_dump_dir=main_folder,
                          sandbox_experiment=None,
                          sandbox_user_token=token,
                          sandbox_user_proxy=FLAGS.proxy(),
                          sandbox_user_signature=FLAGS.signature()):
        running_process = AntTrain(ant_context,
                                  name,
                                  data_factory,
                                  dump_dir,
                                  token,
                                  task,
                                  main_file=main_file,
                                  main_folder=main_folder,
                                  main_param=main_param,
                                  time_stamp=time_stamp,
                                  skip_training=FLAGS.skip_training(),
                                  running_platform=FLAGS.running_platform(),
                                  proxy=FLAGS.proxy(),
                                  signature=FLAGS.signature(),
                                  devices=FLAGS.devices())
        running_process.start()
    except Exception as e:
      print(e)
      traceback.print_exc()
  elif ant_cmd == 'challenge':
    try:
      with running_sandbox(sandbox_dump_dir=main_folder,
                          sandbox_experiment=None,
                          sandbox_user_token=token,
                          sandbox_user_proxy=FLAGS.proxy(),
                          sandbox_user_signature=FLAGS.signature()):
        running_process = AntChallenge(ant_context,
                                      name,
                                      data_factory,
                                      dump_dir,
                                      token,
                                      task,
                                      FLAGS.benchmark(),
                                      main_file=main_file,
                                      main_folder=main_folder,
                                      main_param=main_param,
                                      time_stamp=time_stamp,
                                      running_platform=FLAGS.running_platform(),
                                      devices=FLAGS.devices())
        running_process.start()
    except Exception as e:
      print(e)
      traceback.print_exc()
  elif ant_cmd == "demo":
    try:
      running_process = AntDemo(ant_context,
                                name,
                                dump_dir,
                                token,
                                task,
                                html_template=FLAGS.html_template(),
                                ip=FLAGS.host_ip(),
                                port=FLAGS.host_port(),
                                time_stamp=time_stamp,
                                devices=FLAGS.devices())

      running_process.start()
    except Exception as e:
      print(e)
      traceback.print_exc()
  elif ant_cmd == "predict":
    try:
      running_process = AntBatch(ant_context,
                                name,
                                FLAGS.host_ip(),
                                FLAGS.host_port(),
                                token,
                                data_factory,
                                dump_dir,
                                task,
                                unlabel=FLAGS.unlabel(),
                                devices=FLAGS.devices(),
                                restore_experiment=FLAGS.restore_experiment())
      running_process.start()
    except Exception as e:
      print(e)
      traceback.print_exc()
  elif ant_cmd == "activelearning":
    try:
      running_process = AntActiveLearning(ant_context,
                                          name,
                                          data_factory,
                                          dump_dir,
                                          token,
                                          task,
                                          main_file=main_file,
                                          main_folder=main_folder,
                                          main_param=main_param,
                                          time_stamp=time_stamp)
      running_process.start()
    except Exception as e:
      print(e)
      traceback.print_exc()

  elif ant_cmd == 'dataset':
    try:
      running_process = AntGenerate(ant_context,
                                    name,
                                    data_factory,
                                    dump_dir,
                                    token,
                                    dataset)
      running_process.start()
    except Exception as e:
      print(e)
      traceback.print_exc()
  elif ant_cmd == 'browser':
    try:
      running_process = AntBrowser(ant_context,
                                  name,
                                  token,
                                  FLAGS.host_ip(),
                                  FLAGS.host_port(),
                                  data_factory,
                                  dataset,
                                  dump_dir)
      running_process.start()
    except Exception as e:
      print(e)
      traceback.print_exc()

  # 9.step clear context
  ant_context.wait_until_clear()

if __name__ == '__main__':
  main()



from . import _mask

# Interface for manipulating masks stored in RLE format.
#
# RLE is a simple yet efficient format for storing binary masks. RLE
# first divides a vector (or vectorized image) into a series of piecewise
# constant regions and then for each piece simply stores the length of
# that piece. For example, given M=[0 0 1 1 1 0 1] the RLE counts would
# be [2 3 1 1], or for M=[1 1 1 1 1 1 0] the counts would be [0 6 1]
# (note that the odd counts are always the numbers of zeros). Instead of
# storing the counts directly, additional compression is achieved with a
# variable bitrate representation based on a common scheme called LEB128.
#
# Compression is greatest given large piecewise constant regions.
# Specifically, the size of the RLE is proportional to the number of
# *boundaries* in M (or for an image the number of boundaries in the y
# direction). Assuming fairly simple shapes, the RLE representation is
# O(sqrt(n)) where n is number of pixels in the object. Hence space usage
# is substantially lower, especially for large simple objects (large n).
#
# Many common operations on masks can be computed directly using the RLE
# (without need for decoding). This includes computations such as area,
# union, intersection, etc. All of these operations are linear in the
# size of the RLE, in other words they are O(sqrt(n)) where n is the area
# of the object. Computing these operations on the original mask is O(n).
# Thus, using the RLE can result in substantial computational savings.
#
# The following API functions are defined:
#  encode         - Encode binary masks using RLE.
#  decode         - Decode binary masks encoded via RLE.
#  merge          - Compute union or intersection of encoded masks.
#  iou            - Compute intersection over union between masks.
#  area           - Compute area of encoded masks.
#  toBbox         - Get bounding boxes surrounding encoded masks.
#  frPyObjects    - Convert polygon, bbox, and uncompressed RLE to encoded RLE mask.
#
# Usage:
#  Rs     = encode( masks )
#  masks  = decode( Rs )
#  R      = merge( Rs, intersect=false )
#  o      = iou( dt, gt, iscrowd )
#  a      = area( Rs )
#  bbs    = toBbox( Rs )
#  Rs     = frPyObjects( [pyObjects], h, w )
#
# In the API the following formats are used:
#  Rs      - [dict] Run-length encoding of binary masks
#  R       - dict Run-length encoding of binary mask
#  masks   - [hxwxn] Binary mask(s) (must have type np.ndarray(dtype=uint8) in column-major order)
#  iscrowd - [nx1] list of np.ndarray. 1 indicates corresponding gt image has crowd region to ignore
#  bbs     - [nx4] Bounding box(es) stored as [x y w h]
#  poly    - Polygon stored as [[x1 y1 x2 y2...],[x1 y1 ...],...] (2D list)
#  dt,gt   - May be either bounding boxes or encoded masks
# Both poly and bbs are 0-indexed (bbox=[0 0 1 1] encloses first pixel).
#
# Finally, a note about the intersection over union (iou) computation.
# The standard iou of a ground truth (gt) and detected (dt) object is
#  iou(gt,dt) = area(intersect(gt,dt)) / area(union(gt,dt))
# For "crowd" regions, we use a modified criteria. If a gt object is
# marked as "iscrowd", we allow a dt to match any subregion of the gt.
# Choosing gt' in the crowd gt that best matches the dt can be done using
# gt'=intersect(dt,gt). Since by definition union(gt',dt)=dt, computing
#  iou(gt,dt,iscrowd) = iou(gt',dt) = area(intersect(gt,dt)) / area(dt)
# For crowd gt regions we use this modified criteria above for the iou.
#
# To compile run "python setup.py build_ext --inplace"
# Please do not contact us for help with compiling.
#
# Microsoft COCO Toolbox.      version 2.0
# Data, paper, and tutorials available at:  http://mscoco.org/
# Code written by Piotr Dollar and Tsung-Yi Lin, 2015.
# Licensed under the Simplified BSD License [see coco/license.txt]

iou         = _mask.iou
merge       = _mask.merge
frPyObjects = _mask.frPyObjects

def encode(bimask):
    if len(bimask.shape) == 3:
        return _mask.encode(bimask)
    elif len(bimask.shape) == 2:
        h, w = bimask.shape
        return _mask.encode(bimask.reshape((h, w, 1), order='F'))[0]

def decode(rleObjs):
    if type(rleObjs) == list:
        return _mask.decode(rleObjs)
    else:
        return _mask.decode([rleObjs])[:,:,0]

def area(rleObjs):
    if type(rleObjs) == list:
        return _mask.area(rleObjs)
    else:
        return _mask.area([rleObjs])[0]

def toBbox(rleObjs):
    if type(rleObjs) == list:
        return _mask.toBbox(rleObjs)
    else:
        return _mask.toBbox([rleObjs])[0]


# -*- coding: UTF-8 -*-
# @Time    : 17-6-6
# @File    : core.py
# @Author  : Jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from collections import defaultdict
from functools import total_ordering
from antgo.utils import logger
import itertools
import sys

if sys.version_info[0] == 2:
  def items(dictionary):
    
    """Return a set-like object, a view on key/value pairs of a dict"""
    return dictionary.viewitems()

  def values(dictionary):
    """Return an object providing a view on the values of a dictionary"""
    return dictionary.viewvalues()

  def get_func_name(function, default=None):
    """Return the name of the function, falling back to a default"""
    return getattr(function, 'func_name', default)
else:
  def items(dictionary):
    """Return a set-like object, a view on key/value pairs of a dict"""
    return dictionary.items()

  def values(dictionary):
    """Return an object providing a view on the values of a dictionary"""
    return dictionary.values()

  def get_func_name(function, default=None):
    """Return the name of the function, falling back to a default"""
    return getattr(function, '__name__', default)


class _DIRTY(object):
  """Class definition for the dirty node special value"""

  def __str__(self):
    return 'DIRTY'

  def __repr__(self):
    return '<antgo.dataflow.DIRTY>'

  def __eq__(self, other):
    return self.__class__ == other.__class__

  def __ne__(self, other):
    return self.__class__ != other.__class__


DIRTY = _DIRTY()


class BaseNode(object):
  """Base class for Inputs and Nodes"""

  _name_counters = defaultdict(int)

  def __init__(self, name=None, value=DIRTY, auto_trigger=False):
    self.name = name or self._generate_name()
    self._value = value
    self._dependents = set()
    self._auto_trigger = auto_trigger

  def _connect(self, dependent):
    """Set the given Node as a dependent of this Node or Input

    Immediately paints the new dependent Node dirty if this Node has
    already been evaluated or if a value has already been set for this
    Input.

    Connecting Nodes always invalidates the triggered Nodes cache.

    """
    if dependent not in self._dependents:
        self._dependents.add(dependent)
        if self._value is not DIRTY:
            dependent._set_value(DIRTY)

  def _disconnect(self, dependent):
    """Remove given Node from the set of dependents of this Node or Input

    Immediately paints the new dependent Node dirty if this Node has
    previously been evaluated or if a value has previously been set for
    this Input.

    Disconnecting Nodes always invalidates the triggered nodes cache.

    """
    if dependent in self._dependents:
      self._dependents.remove(dependent)
      if self._value is not DIRTY:
        dependent._set_value(DIRTY)

  def _set_value(self, value):
    """Set a new value for this Node or Input

    If this caused the value to change, paints dependent Nodes dirty and
    returns the set of those dependent Nodes which are marked "triggered"
    and should be re-evaluated.

    When called by ``set_value`` from external code, the ``get_triggered``
    argument must be ``True`` so the return value is cached.  Internal
    calls set ``get_triggered=False`` so memory isn't wasted for caching the
    triggered dependents of intermediate Nodes.

    This private method can be used as a debugging tool to set values of
    non-input Nodes.

    """
    # test if neither, one of or both the old and the new value are DIRTY
    dirty_count = len([v for v in [value, self._value] if v is DIRTY])
    if dirty_count == 2:
      # both DIRTY, no need to touch anything
      return set()
    if dirty_count == 0 and self._value_eq(value):
      # both non-DIRTY but equal, no need to touch anything
      return set()
    # either one is DIRTY, or values aren't equal, update the value and
    # paint the dependent Nodes dirty
    self._value = value
    self._set_dependents_dirty()
    # try to trigger some auto-node
    if DIRTY != value:
      self._trigger()

  def _value_eq(self, other_value):
    return False

  def get_value(self):
    """Return the value of the object"""
    raise NotImplementedError('The get_value() method must be defined '
                              'for subclasses of BaseNode')

  def _set_dependents_dirty(self):
    """Paint all dependent Nodes dirty

    Paints direct dependent Nodes dirty, which causes recursive painting
    for the whole dependent Nodes tree.

    """
    for dependent in self._dependents:
        dependent._set_value(DIRTY)

  def _trigger(self):
    if self._auto_trigger:
      self.get_value()
    for dependent in self._dependents:
      dependent._trigger()

  def _force_inputs_dirty(self):
    self._value = DIRTY

  def _reset_iteration_state(self):
    pass

  def _generate_name(self):
    """Generate a unique name for this Node or Input object

    The name includes:

    * the name of the class
    * an auto-incremented number

    """
    counters = self._name_counters
    template = '{class_name}-{counter}'
    counters[self.__class__] += 1
    return template.format(class_name=self.__class__.__name__,
                           counter=counters[self.__class__])

  def __repr__(self):
    return ('<{self.__class__.__name__} {self.name}: {self._value}>'
            .format(self=self))


class Input(BaseNode):
  """The input node class for reactive programming

  Constructor arguments
  ---------------------

  name (optional): string
          The internal name of the Input. Used in the
          ``__repr__`` of the object. If omitted, a name is
          automatically generated.

  value (optional):
          The initial value for the Input.

  Examples of Inputs::

      >>> input_1 = Input()  # no name, no default value
      >>> input_2 = Input(value=10.0)  # input with a default value
      >>> exponent = Input(name='exponent')  # named input
      >>> sensor = Input(name='sensor', value=-5.3)  # named, with default

  """
  def __init__(self, name):
    super(Input, self).__init__(name)

  def get_value(self):
    return self._value

  def set_value(self, new_value):
    """Set a new value for an Input

    If this caused the value to change, paints dependent Nodes dirty and
    returns the set of those dependent Nodes which are marked "triggered"
    and should be re-evaluated.

    """
    return self._set_value(new_value)

  value = property(get_value, set_value)


@total_ordering
class Node(BaseNode):
  """The Node class for reactive programming

  Constructor arguments
  ---------------------

  name (optional): string
          The internal name of the Node. Used in the ``__repr__`` of the
          object. If omitted, a name is automatically generated.

  action: callable(*positional_inputs, **keyword_inputs)
          The function for calculating the value of a calculated node.
          Values from inputs are provided in positional and keyword arguments
          as defined in the ``inputs=`` argument.

  inputs (optional): ((Input/Node, ...), {key: Input/Node, ...})
          The Nodes and Inputs whose values are used as inputs for the
          action.  This argument can be created with ``Node.inputs()`` which
          provides a cleaner syntax.

  triggered: boolean (default=False)
          ``True`` is this Node shoud be automatically evaluated when any of
          its dependency Nodes or Inputs change value

  Examples of Nodes::

      >>> input_1, input_2, exponent = [Input() for i in range(3)]
      >>> # sum Node with two positional inputs
      >>> sum_node = Node(action=lambda *args: sum(args),
      ...                 inputs=Node.inputs(input_1, input_2))
      >>> # triggered (auto-calculated) Node with two keyword inputs
      >>> triggered_node = Node(
      ...     action=lambda a, x: a ** x,
      ...     inputs=Node.inputs(a=input_1, x=exponent))

  """
  def __init__(self,
               name=None,
               action=None,
               inputs=((), None), auto_trigger=False):
    self._action = action  # must be set before generating name
    super(Node, self).__init__(name, value=DIRTY, auto_trigger=auto_trigger)
    self._positional_inputs = ()
    self._keyword_inputs = {}
    if inputs is not None:
      self.set_inputs(*inputs[0], **inputs[1] or {})
    self._set_dependents_dirty()
    self._iteration_reset_state = False

  def _evaluate(self):
    """Calculate the value for the Node

    Calls the action of the Node using values from the inputs of the Node.
    Returns the result of the action function.

    This function can also be overridden in subclasses if a class-based
    approach to creating Node actions is preferred.

    """
    if not self._action:
      raise NotImplementedError('You must define the action= argument '
                                  'when instantiating the Node')
    try:
      positional_values = [i.get_value()
                           for i in self._positional_inputs]
      keyword_values = {name: i.get_value()
                        for name, i in items(self._keyword_inputs)}

      # clear reset state
      self._iteration_reset_state = False
    except StopIteration:
      # reset all input
      self._reset_iteration_state()
      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s'%(info[0], info[1]))
      exit(-1)

    value = self._action(*positional_values, **keyword_values)
    # if (getattr(self._action, 'output_type', None) is not None):
    #     # Output type checking has been enabled, and the node's action
    #     # does specify the expected output type. Check that the calculated
    #     # value matches that type.
    #     self._verify_output_type(value)
    return value

  @staticmethod
  def inputs(*args, **kwargs):
    """Construct a value for the inputs= kwarg of the constructor

    Allows writing this::

        >>> inputs = [Input() for i in range(4)]
        >>> node = Node(inputs=Node.inputs(inputs[0], inputs[1],
        ...                                kw1=inputs[2], kw2=inputs[3]))

    instead of this::

        >>> node = Node(inputs=([inputs[0], inputs[1]],
        ...                     {'kw1': inputs[2], 'kw2': inputs[3]}))

    """
    return args, kwargs

  def output(self, index=0):
    return [self], {}

  def _verify_output_type(self, value):
    """Assert that the given value matches the action's output type

    This check should be run only in development if the developer wants to
    ensure the consistency of a graph's types.

    This method may only be called if the node's action has a non-None
    ``output_type`` attribute.

    Arguments
    ---------
    value: The value whose type is to be checked

    Raises
    ------
    TypeError: The value doesn't match the desired output type of the
               node's action

    """
    if not isinstance(value, self._action.output_type):
      raise TypeError(
          "The output value type {value_type!r} for [{self.name}]\n"
          "doesn't match the expected type "
          '{self._action.output_type.__name__!r} for action '
          '"{self._action.name}".'
          .format(value_type=type(value).__name__, self=self))

  def set_inputs(self, *args, **kwargs):
    """Replace current positional and keyword inputs"""
    for inp in self._iterate_inputs():
        inp._disconnect(self)
    self._positional_inputs = args
    self._keyword_inputs = kwargs
    for inp in self._iterate_inputs():
        inp._connect(self)

  def get_value(self):
    """Return Node value, evaluate if needed and paint dependents dirty"""
    if DIRTY == self._value:
      self._value = self._evaluate()
      #self._set_dependents_dirty()

    return self._value

  def is_dirty(self):
    return DIRTY == self._value

  def iterator_value(self):
    while True:
      try:
        self._value = self._evaluate()
        yield self._value
        self._force_inputs_dirty()
      except StopIteration:
        return
      except:
        logger.error('data flow pipeline error')
        return

  def set_value(self, new_value):
    """Set a new value for an Input

    If this caused the value to change, paints dependent Nodes dirty and
    returns the set of those dependent Nodes which are marked "triggered"
    and should be re-evaluated.

    """
    return self._set_value(new_value)

  value = property(get_value, set_value)

  def _iterate_inputs(self):
    """Iterate through positional and keyword inputs"""
    return itertools.chain(self._positional_inputs,
                           values(self._keyword_inputs))

  def _generate_name(self):
    """Generate a unique name for this Node object

    The name includes:

    * the name of the node class
    * the function name of the ``_action`` if it's defined
      and isn't a lambda
    * an auto-incremented number

    """
    action_name = get_func_name(self._action, '<lambda>')
    if action_name == '<lambda>':
      return super(Node, self)._generate_name()
    counters = self._name_counters
    counters[self.__class__, action_name] += 1
    template = '{class_name}-{action_name}-{counter}'
    return template.format(class_name=self.__class__.__name__,
                           action_name=action_name,
                           counter=counters[self.__class__, action_name])

  def _force_inputs_dirty(self):
    if DIRTY == self._value:
      return
    
    for i in self._positional_inputs:
      i._force_inputs_dirty()
    for name, i in items(self._keyword_inputs):
      i._force_inputs_dirty()
    self._value = DIRTY

  def _reset_iteration_state(self):
    if self._iteration_reset_state:
        return
    for i in self._positional_inputs:
        i._reset_iteration_state()
    for name, i in items(self._keyword_inputs):
        i._reset_iteration_state()
    self._iteration_reset_state = True

  def __lt__(self, other):
    return self.name < other.name




# -*- coding: UTF-8 -*-
# File: svhn.py
# Author: jian(jian@mltalker.com)
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
import random
import numpy as np
from six.moves import range
import copy
from antgo.utils import logger, get_rng
from antgo.dataflow.dataset import Dataset
from antgo.utils.fs import download
from antgo.utils.fs import maybe_here
import scipy.io
__all__ = ['SVHNDigit']

SVHN_URL = "http://ufldl.stanford.edu/housenumbers/"
class SVHNDigit(Dataset):
  def __init__(self, name, dir=None,params=None):
    """
    :param name: 'train', 'test'
    :param data_dir: a directory containing the original {train,test}_32x32.mat
    """
    super(SVHNDigit, self).__init__(name, dir, params)

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    if name == "val":
        super(SVHNDigit,self).__init__("val",dir)
    else:
        super(SVHNDigit, self).__init__(name, dir)

    filename = ""
    if self.train_or_test == 'val':
      if not os.path.exists(os.path.join(self.dir, 'extra_32x32.mat')):
        self.download(self.dir, ['extra_32x32.mat'], default_url=SVHN_URL)

      filename = os.path.join(self.dir, 'extra' + '_32x32.mat')
    else:
      if not os.path.exists(os.path.join(self.dir, self.train_or_test + '_32x32.mat')):
        self.download(self.dir, [self.train_or_test + '_32x32.mat'], default_url=SVHN_URL)

      filename = os.path.join(self.dir, self.train_or_test + '_32x32.mat')
    assert os.path.isfile(filename), \
      "File {} not found! Please download it from {}.".format(filename, SVHN_URL)

    logger.info("Loading {} ...".format(filename))
    data = scipy.io.loadmat(filename)
    self.X = data['X'].transpose(3, 0, 1, 2)
    self.Y = data['y'].reshape((-1))
    self.Y[np.where(self.Y == 10)] = 0
    self.Y = self.Y.astype(np.uint8)

    self.ids = list(range(self.Y.shape[0]))

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')

    svhn_val = SVHNDigit('val', self.dir)
    return self, svhn_val

  @property
  def size(self):
    return len(self.ids)

  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      idxs = copy.deepcopy(self.ids)
      if self.rng:
          self.rng.shuffle(idxs)

      for k in idxs:
          yield [self.X[k], self.Y[k]]

  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    return [self.X[self.ids[id]], self.Y[self.ids[id]]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: argscope.py
# Author: Yuxin Wu <ppwwyyxxc@gmail.com>
from contextlib import contextmanager
from collections import defaultdict
import inspect
import copy
import six

__all__ = ['argscope', 'get_arg_scope']

_ArgScopeStack = []

@contextmanager
def argscope(layers, **param):
    if not isinstance(layers, list):
        layers = [layers]

    def _check_args_exist(l):
        args = inspect.getargspec(l).args
        for k, v in six.iteritems(param):
            assert k in args, "No argument {} in {}".format(k, l.__name__)

    for l in layers:
        assert hasattr(l, 'f'), "{} is not a registered layer".format(l.__name__)
        _check_args_exist(l.f)

    new_scope = copy.copy(get_arg_scope())
    for l in layers:
        new_scope[l.__name__].update(param)
    _ArgScopeStack.append(new_scope)
    yield
    del _ArgScopeStack[-1]

def get_arg_scope():
    """
    :returns: the current argscope.
        An argscope is a dict of dict: dict[layername] = {arg: val}
    """
    if len(_ArgScopeStack) > 0:
        return _ArgScopeStack[-1]
    else:
        return defaultdict(dict)


# -*- coding: UTF-8 -*-
# @Time    : 17-12-14
# @File    : simpleimages.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import copy
import multiprocessing
import six
import sys
from antgo.utils.shared_queue import SharedQueue
import numpy as np


class EndSignal():
    pass


class MultiprocessReader(object):
    def __init__(self, dataset, 
                        transformers=None,
                        num_workers=4,
                        buffer_size=128,
                        batch_size=0,
                        drop_last=True,
                        daemon=True):
        super().__init__()

        self.queue = SharedQueue(buffer_size, memsize=1024**3)

        self.num_workers = num_workers
        self.transformers = transformers
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.dataset = dataset
        self.daemon = daemon
        
    def _read_into_queue(self, samples, dataset, queue):
        end = EndSignal()
        try:
            for sample in samples:
                if sample is None:
                    raise ValueError("Sample has None.s")
                data = dataset.at(sample)
                if self.transformers is not None:
                    for t in self.transformers:
                        data = t.action(data)
                queue.put(data)

            queue.put(end)
        except:
            queue.put("")
            six.reraise(*sys.exc_info())

    def iterator_value(self):
        total_samples = [[] for i in range(self.num_workers)]
        for i, sample_i in enumerate(range(self.dataset.size)):
            index = i % self.num_workers
            total_samples[index].append(sample_i)

        for i in range(self.num_workers):
            p = multiprocessing.Process(
                target=self._read_into_queue, args=(total_samples[i], self.dataset, self.queue))
            if self.daemon:
                p.daemon = self.daemon
            p.start()

        finish_num = 0
        batch_data = [],[]
        while finish_num < self.num_workers:
            sample = self.queue.get()
            if isinstance(sample, EndSignal):
                finish_num += 1
            elif sample == "":
                raise ValueError("Multiprocess reader raises an exception.")
            else:
                if self.batch_size == 0:
                    yield sample
                else:
                    batch_data[0].append(sample[0])
                    batch_data[1].append(sample[1])
                    if len(batch_data[0]) == self.batch_size:
                        a = np.stack(batch_data[0], 0)
                        yield a, batch_data[1]
                        batch_data = [],[]

        if self.batch_size > 0:
            if len(batch_data[0]) != 0 and not self.drop_last:
                a = np.stack(batch_data[0],0)
                yield a, batch_data[1]
                batch_data = [],[]



# -*- coding: UTF-8 -*-
# File: pascal_voc.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals

import os, sys
import numpy as np
import random
import six
from six.moves import urllib, range
import copy
import logging
import tarfile
import xml.etree.ElementTree as ET
import scipy.sparse
from antgo.utils.fs import download
from antgo.utils.fs import maybe_here_match_format
from antgo.utils import logger, get_rng
from antgo.dataflow.dataset.dataset import *

__all__ = ['Pascal2007', 'Pascal2012']

PASCAL2007_URL="http://host.robots.ox.ac.uk/pascal/VOC/voc2007"
PASCAL2012_URL="http://host.robots.ox.ac.uk/pascal/VOC/voc2012"


class PascalBase(Dataset):
  def __init__(self, year, image_set, dir=None, ext_params=None):
    super(PascalBase, self).__init__(image_set, dir, ext_params)
    self._year = year
    self._image_set = image_set
    self._devkit_path = dir

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self._image_index = self.load_samples()
      return

    if self._year == '2007':
      if not os.path.exists(os.path.join(self._devkit_path, 'VOCdevkit')):
        self.download(self.dir, ['VOCtrainval_06-Nov-2007.tar', 'VOCtest_06-Nov-2007.tar'], default_url=PASCAL2007_URL)
      maybe_data_path = maybe_here_match_format(self._devkit_path, 'VOC' + self._year)
      if maybe_data_path is None:
        # auto untar
        tar = tarfile.open(os.path.join(self.dir,'VOCtrainval_06-Nov-2007.tar'), 'r')
        tar.extractall(self.dir)
        tar.close()

        tar = tarfile.open(os.path.join(self.dir, 'VOCtest_06-Nov-2007.tar'), 'r')
        tar.extractall(self.dir)
        tar.close()
    else:
      if not os.path.exists(os.path.join(self._devkit_path, 'VOCdevkit')):
        self.download(self.dir, ['VOCtrainval_11-May-2012.tar'], default_url=PASCAL2012_URL)
        self.download(self.dir, ['VOC2012test.tar'], default_url='http://host.robots.ox.ac.uk:8080/eval/downloads/')

      maybe_data_path = maybe_here_match_format(self._devkit_path, 'VOC' + self._year)
      if maybe_data_path is None:
        # auto untar
        tar = tarfile.open(os.path.join(self.dir, 'VOCtrainval_11-May-2012.tar'), 'r')
        tar.extractall(self.dir)
        tar.close()

        tar = tarfile.open(os.path.join(self.dir, 'VOC2012test.tar'), 'r')
        tar.extractall(self.dir)
        tar.close()

    self._data_path = os.path.join(self.dir,'VOCdevkit', 'VOC' + self._year)
    self._classes = ('background',
                     'aeroplane',
                     'bicycle',
                     'bird',
                     'boat',
                     'bottle',
                     'bus',
                     'car',
                     'cat',
                     'chair',
                     'cow',
                     'diningtable',
                     'dog',
                     'horse',
                     'motorbike',
                     'person',
                     'pottedplant',
                     'sheep',
                     'sofa',
                     'train',
                     'tvmonitor')

    self._c_2_rgb = {0: (0, 0, 0),
                     1: (128, 0, 0),
                     2: (0, 128, 0),
                     3: (128, 128, 0),
                     4: (0, 0, 128),
                     5: (128, 0, 128),
                     6: (0, 128, 128),
                     7: (128, 128, 128),
                     8: (64, 0, 0),
                     9: (192, 0, 0),
                     10: (64, 128, 0),
                     11: (192, 128, 0),
                     12: (64, 0, 128),
                     13: (192, 0, 128),
                     14: (64, 128, 128),
                     15: (192, 128, 128),
                     16: (0, 64, 0),
                     17: (128, 64, 0),
                     18: (0, 192, 0),
                     19: (128, 192, 0),
                     20: (0, 64, 128)}
    self._c_2_rgb_index = {}
    for k, v in self._c_2_rgb.items():
      self._c_2_rgb_index[k] = v[0]*255*255+v[1]*255+v[2]

    self._action_classes = {'phoning': 1,
                            'playinginstrument': 2,
                            'reading': 3,
                            'ridingbike': 4,
                            'ridinghorse': 5,
                            'running': 6,
                            'takingphoto': 7,
                            'usingcomputer': 8,
                            'walking': 9,
                            'jumping': 10,
                            'other': 0}

    self._num_classes = len(self._classes)
    self._class_to_ind = dict(zip(self._classes, range(self._num_classes)))
    self._image_ext = '.jpg'
    self._image_index = self._load_image_set_index()

    # PASCAL specific config options
    self.config = {'cleanup': True,
                   'use_salt': True,
                   'use_diff': False,
                   'matlab_eval': False,
                   'rpn_file': None,
                   'min_size': 2}

    assert os.path.exists(self._data_path), 'path does not exist: {}'.format(self._data_path)

    self._aug_dataset = []
    self._aug_seg = []
    if self._year == '2012':
      train_aug = getattr(self, 'aug', None)
      task_type = getattr(self, 'task_type', None)
      if train_aug is not None and task_type == 'SEGMENTATION':
        assert(os.path.exists(os.path.join(self.dir, 'SegmentationClassAug')))
        assert(os.path.exists(os.path.join(self.dir, 'trainaug.txt')))
        with open(os.path.join(self.dir, 'trainaug.txt'), 'r') as fp:
          content = fp.readline()

          while(content):
            content = content.strip()
            if content == "":
              break

            self._aug_dataset.append(os.path.join(self.dir, 'VOCdevkit/VOC2012/JPEGImages', '%s.jpg'%content))
            self._aug_seg.append(os.path.join(self.dir, 'SegmentationClassAug', '%s.png'%content))
            content = fp.readline()

  @property
  def size(self):
    return len(self._image_index) + len(self._aug_dataset)

  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = list(range(self.size))
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      # data index shuffle
      idxs = np.arange(len(self._image_index))
      if self.rng:
        self.rng.shuffle(idxs)

      for k in idxs:
        # real index
        index = self._image_index[k]
        if self.train_or_test == 'test' and self._year == '2012':
          image = imread(self.image_path_from_index(index))
          yield [image, {'file_id': index + self._image_ext}]

          continue

        # annotation
        gt_roidb = self._load_roidb(index)

        task_type = getattr(self, 'task_type', None)
        if task_type is not None:
          if task_type == 'ACTION_CLASSIFICATION':
            keep_index = []
            for obj_index, obj_name in enumerate(gt_roidb['category']):
              if obj_name == 'person':
                keep_index.append(obj_index)

            if len(keep_index) == 0:
              continue

            keep_index = [ki for ki in keep_index if gt_roidb['person_action'][ki] is not None]
            if len(keep_index) == 0:
              continue

            gt_roidb['category_id'] = np.concatenate([gt_roidb['person_action'][ki].reshape((1,-1)) for ki in keep_index])
            gt_roidb['bbox'] = gt_roidb['bbox'][keep_index,:]
            gt_roidb['area'] = gt_roidb['area'][keep_index]
            gt_roidb['difficult'] = [gt_roidb['difficult'][ki] for ki in keep_index]

            gt_roidb.pop('person_action')

        # image
        image = imread(self.image_path_from_index(index))
        # image original size
        gt_roidb['info'] = (image.shape[0], image.shape[1], image.shape[2])
        gt_roidb['id'] = k
        
        # [img, groundtruth]
        gt_roidb.update({'file_id': index})
        yield [image, gt_roidb]

      if len(self._aug_dataset) > 0:
        aug_indxs = np.arange(len(self._aug_dataset))
        if self.rng:
          self.rng.shuffle(aug_indxs)

        for k in aug_indxs:
          image_path = self._aug_dataset[k]
          image_seg_path = self._aug_seg[k]

          image = imread(image_path)
          seg_img = imread(image_seg_path)
          if len(seg_img.shape) == 3:
            seg_img = seg_img[:,:,0]

          seg_img[np.where(seg_img == 255)] = 0

          yield image, {'segmentation_map': seg_img,
                        'id': k+len(self._image_index),
                        'info': (image.shape[0], image.shape[1], image.shape[2])}

  
  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    index = self._image_index[id]
    gt_roidb = self._load_roidb(index)

    task_type = getattr(self, 'task_type', None)
    if task_type is not None:
      if task_type == 'ACTION_CLASSIFICATION':
        keep_index = []
        for obj_index, obj_name in enumerate(gt_roidb['category']):
          if obj_name == 'person':
            keep_index.append(obj_index)

        if len(keep_index) == 0:
          return [None, None]

        keep_index = [ki for ki in keep_index if gt_roidb['person_action'][ki] is not None]
        if len(keep_index) == 0:
          return [None, None]

        gt_roidb['category_id'] = np.concatenate([gt_roidb['person_action'][ki].reshape((1, -1)) for ki in keep_index])
        gt_roidb['bbox'] = gt_roidb['bbox'][keep_index, :]
        gt_roidb['area'] = gt_roidb['area'][keep_index]
        gt_roidb['difficult'] = [gt_roidb['difficult'][ki] for ki in keep_index]

        gt_roidb.pop('person_action')

    image = imread(self.image_path_from_index(index))
    gt_roidb['info'] = (image.shape[0], image.shape[1], image.shape[2])
    gt_roidb['id'] = id

    # [img,groundtruth]
    return [image, gt_roidb]
  
  def image_path_from_index(self, index):
    """
    Construct an image path from the image's "index" identifier.
    """
    image_path = os.path.join(self._data_path, 'JPEGImages',
                              index + self._image_ext)
    assert os.path.exists(image_path), \
            'Path does not exist: {}'.format(image_path)
    return image_path
    
  def image_path_at(self, i):
    """
    Return the absolute path to image i in the image sequence.
    """
    return self.image_path_from_index(self._image_index[i])

  def _load_roidb(self, index):
    return self._load_pascal_annotation(index)
    
  def _load_image_set_index(self):
    """
    Load the indexes listed in this dataset's image set file.
    """
    # Example path to image set file:
    # self._devkit_path + /VOCdevkit2007/VOC2007/ImageSets/Main/val.txt
    image_set_file = os.path.join(self._data_path, 'ImageSets', 'Main',
                                  self._image_set + '.txt')

    task_type = getattr(self, 'task_type', None)
    if task_type is not None:
      if task_type == 'ACTION_CLASSIFICATION':
        image_set_file = os.path.join(self._data_path, 'ImageSets', 'Action',
                                  self._image_set + '.txt')
      elif task_type == 'SEGMENTATION':
        image_set_file = os.path.join(self._data_path, 'ImageSets', 'Segmentation',
                                  self._image_set + '.txt')

    assert os.path.exists(image_set_file), \
            'Path Does not Exist: {}'.format(image_set_file)
    with open(image_set_file) as f:
        image_index = [x.strip() for x in f.readlines()]
    return image_index
    
  def _load_pascal_annotation(self, index):
    """
    Load image and bounding boxes info from XML file in the PASCAL VOC
    format.
    """
    filename = os.path.join(self._data_path, 'Annotations', index + '.xml')
    tree = ET.parse(filename)
    objs = tree.findall('object')

    # Exclude the samples labeled as difficult
    # non_diff_objs = [
    #     obj for obj in objs if int(obj.find('difficult').text) == 0]
    # if len(non_diff_objs) != len(objs):
    #     print 'Removed {} difficult objects'.format(
    #         len(objs) - len(non_diff_objs))
    # objs = non_diff_objs

    num_objs = len(objs)

    boxes = np.zeros((num_objs, 4), dtype=np.uint16)
    category_id = np.zeros((num_objs), dtype=np.int32)
    # "Seg" area for pascal is just the box area
    area = np.zeros((num_objs), dtype=np.float32)
    category = []
    difficult = []

    segmented = tree.find('segmented')
    has_seg = 0
    if segmented is not None:
      has_seg = int(segmented.text)

    seg_img = None
    if has_seg:
      seg_file = os.path.join(self._data_path, 'SegmentationClass', index + '.png')
      seg_img = imread(seg_file)

    # Load object bb and segmentation into a data frame.
    segmentation = []
    segmentation_map = None
    person_action = []
    for ix, obj in enumerate(objs):
      bbox = obj.find('bndbox')
      # Make pixel indexes 0-based
      x1 = float(bbox.find('xmin').text) - 1
      y1 = float(bbox.find('ymin').text) - 1
      x2 = float(bbox.find('xmax').text) - 1
      y2 = float(bbox.find('ymax').text) - 1
      cls = self._class_to_ind[obj.find('name').text.lower().strip()]
      boxes[ix, :] = [x1, y1, x2, y2]
      category_id[ix] = cls
      area[ix] = (x2 - x1 + 1) * (y2 - y1 + 1)
      category.append(obj.find('name').text.lower().strip())
      difficult.append(int(obj.find('difficult').text))

      if has_seg and seg_img is not None:
        obj_seg = np.zeros((seg_img.shape[0], seg_img.shape[1]), np.uint8)
        seg_img_int = seg_img.astype(np.int32)
        seg_img_index = seg_img_int[:,:,0]*255*255 + seg_img_int[:,:,1]*255 + seg_img_int[:,:,2]

        obj_seg[np.where(seg_img_index == self._c_2_rgb_index[cls])] = cls
        segmentation.append(obj_seg)
        if segmentation_map is None:
          segmentation_map = np.zeros((seg_img.shape[0], seg_img.shape[1]), np.uint8)

        segmentation_map[np.where(seg_img_index == self._c_2_rgb_index[cls])] = cls

      if category[-1] == 'person':
        action_cls = obj.find('actions')
        if action_cls is not None:
          action_label = np.zeros((11), np.int32)
          for action_cc in action_cls.getchildren():
            action_label[self._action_classes[action_cc.tag]] = int(action_cc.text)

          person_action.append(action_label)
        else:
          person_action.append(None)
      else:
        person_action.append(None)

    annotation = {'bbox': boxes,
                  'category_id': category_id,
                  'category': category,
                  'flipped': False,
                  'difficult': difficult,
                  'area': area,
                  'person_action': person_action}

    if has_seg:
      annotation.update({'segmentation': segmentation, 'segmentation_map': segmentation_map})

    return annotation


class Pascal2007(PascalBase):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    super(Pascal2007,self).__init__('2007', train_or_test, dir, ext_params)

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')
    validation_pascal2007 = Pascal2007('val', self.dir)

    return self, validation_pascal2007


class Pascal2012(PascalBase):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    super(Pascal2012,self).__init__('2012', train_or_test, dir, ext_params)

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')
    validation_pascal2012 = Pascal2012('val', self.dir)

    return self, validation_pascal2012


# -*- coding: UTF-8 -*-
# @Time    : 17-12-6
# @File    : tftools.py
# @Author  : Jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import tensorflow as tf
from tensorflow.python.tools import freeze_graph
from tensorflow.python.tools import optimize_for_inference_lib
from tensorflow.python.platform import gfile
from antgo.ant.utils import *
slim = tf.contrib.slim
from antgo.utils import logger
from antgo.dataflow.dataset.empty_dataset import *
from antgo.ant import flags
from datetime import datetime
from multiprocessing import Process
import yaml
FLAGS = flags.AntFLAGS


def tftool_frozen_graph(ctx,
                        dump_dir,
                        time_stamp,
                        input_node_names,
                        output_node_names):
  # 1.step build inference model
  now_time_stamp = datetime.fromtimestamp(time_stamp).strftime('%Y%m%d.%H%M%S.%f')
  dump_dir = os.path.join(dump_dir, now_time_stamp)
  if not os.path.exists(dump_dir):
    os.makedirs(dump_dir)

  ablation_blocks = getattr(ctx.params, 'ablation', [])
  for b in ablation_blocks:
    ctx.deactivate_block(b)

  # independent process
  # ctx.call_frozen_process(dump_dir)
  logger.info('deploy infer model')
  p = multiprocessing.Process(target=ctx.call_infer_process, args=(None, dump_dir))
  p.start()
  p.join()

  # 2.step check input_node_names and output_node_names
  if input_node_names == '' or output_node_names == '':
    logger.error('must set input_nodes and output_nodes')
    return

  # 3.step check infer_graph.pbtxt in experiment
  if not os.path.exists(os.path.join(dump_dir, 'infer_graph.pbtxt')):
    logger.error('dont exist graph file')
    return

  # 4.step build frozen graph
  # We retrieve our checkpoint fullpath
  checkpoint = tf.train.get_checkpoint_state(dump_dir)
  input_checkpoint = checkpoint.model_checkpoint_path

  # We precise the file fullname of our freezed graph
  input_checkpoint_path = os.path.join(dump_dir, input_checkpoint.split('/')[-1])

  input_graph_path = os.path.join(dump_dir, 'infer_graph.pbtxt')
  output_graph_path = os.path.join(dump_dir, "frozen_model.pb")

  input_saver_path = ""
  input_binary = False

  restore_op_name = 'save/restore_all'
  filename_tensor_name = 'save/Const:0'

  clear_devices = True
  initializer_nodes = ""

  freeze_graph.freeze_graph(input_graph_path,
                            input_saver_path,
                            input_binary,
                            input_checkpoint_path,
                            output_node_names,
                            restore_op_name,
                            filename_tensor_name,
                            output_graph_path,
                            clear_devices, initializer_nodes)
  
  # 5.step build optimized graph
  with tf.Graph().as_default() as graph:
    input_graph_def = graph.as_graph_def()
    with tf.gfile.FastGFile(output_graph_path,'rb') as f:
      input_graph_def.ParseFromString(f.read())

      _ = tf.import_graph_def(input_graph_def, name="")
      summary_write = tf.summary.FileWriter(dump_dir, graph)


  output_graph_def = optimize_for_inference_lib.optimize_for_inference(input_graph_def,
                                                                       input_node_names.split(','),
                                                                       output_node_names.split(','),
                                                                       tf.float32.as_datatype_enum)

  optimized_path = os.path.join(dump_dir, "optimized_model.pb")
  f = tf.gfile.FastGFile(optimized_path, "wb")
  f.write(output_graph_def.SerializeToString())


def tftool_visualize_pb(pb_path):
  if not os.path.exists(pb_path):
    logger.error('pb model file dont exist')
    return

  logger.info('load model pb')
  graph = tf.get_default_graph()
  graphdef = graph.as_graph_def()
  graphdef.ParseFromString(gfile.FastGFile(pb_path, "rb").read())

  _ = tf.import_graph_def(graphdef, name="")

  logger.info('start model FLOPs statistic')
  flops = tf.profiler.profile(graph, options=tf.profiler.ProfileOptionBuilder.float_operation())
  logger.info('model FLOPs: {}'.format(flops.total_float_ops))

  logger.info('generate visualization data')
  summary_write = tf.summary.FileWriter("./", graph)
  logger.info('open tensorboard --logdir=.')


def _int64_feature(values):
  if not isinstance(values, (tuple, list)):
    values = [values]
  return tf.train.Feature(int64_list=tf.train.Int64List(value=values))


def _bytes_feature(values):
  return tf.train.Feature(bytes_list=tf.train.BytesList(value=[values]))


def _get_dataset_record_filename(dataset_record_dir, dataset_name, split_name, shard_id, num_shards):
  output_filename = '%s_%s_%05d-of-%05d.tfrecord' % (dataset_name, split_name, shard_id, num_shards)
  return os.path.join(dataset_record_dir, output_filename)


def tftool_generate_image_records(data_dir,
                                  label_dir,
                                  record_dir,
                                  label_flag='_label',
                                  train_or_test='train',
                                  num_shards=20,
                                  dataset_flag='antgo'):
  



  # 1.step search all files in data_dir
  data_files = os.listdir(data_dir)
  is_jpg = False
  is_png = False
  kp_maps = {}
  for f in data_files:
    if f.split('.')[-1].lower() in ['jpg', 'png', 'jpeg']:
      if not is_jpg and not is_png:
        if f.split('.')[-1].lower() in ['jpg', 'jpeg']:
          is_jpg = True
        else:
          is_png = True
      f_path = os.path.join(data_dir, f)
      kp_maps[f.split('.')[0]] = f_path
  
  label_files = os.listdir(label_dir)
  
  prepare_data_files = []
  prepare_label_files = []
  for f in label_files:
    prefix = f.split('.')[0].replace(label_flag, '')
    if prefix in kp_maps:
      prepare_data_files.append(kp_maps[prefix])
      prepare_label_files.append(os.path.join(label_dir, f))
    else:
      sys.stdout.write('%f couldnt find data - label pair'% f)
      sys.stdout.flush()

  sys.stdout.write('total files %d are splited to %d shards'%(len(prepare_data_files),int(num_shards)))
  sys.stdout.flush()

  # 2.step transform to tfrecord
  num_per_shard = int(np.math.ceil(len(prepare_data_files) / float(num_shards)))
  with tf.Session('') as sess:
    for shard_id in range(num_shards):
      record_filename = _get_dataset_record_filename(record_dir, dataset_flag, train_or_test, shard_id, num_shards)

      with tf.python_io.TFRecordWriter(record_filename) as tfrecord_writer:
        start_ndx = shard_id * num_per_shard
        end_ndx = min((shard_id + 1) * num_per_shard, len(prepare_data_files))

        for i in range(start_ndx, end_ndx):
          a_data = tf.gfile.FastGFile(prepare_data_files[i], 'rb').read()
          b_data = tf.gfile.FastGFile(prepare_label_files[i], 'rb').read()
          
          sys.stdout.write('\r>> Converting image %d/%d shard %d\n' % (i + 1, len(prepare_data_files), shard_id))
          sys.stdout.flush()

          format_str = 'jpg' if is_jpg else 'png'
          example = tf.train.Example(features=
                                      tf.train.Features(feature={
                                                                'image/encoded': _bytes_feature(a_data),
                                                                'image/format': _bytes_feature(format_str.encode('utf-8')),
                                                                'label/encoded': _bytes_feature(b_data),
                                                                'label/format': _bytes_feature(format_str.encode('utf-8'))}))
          tfrecord_writer.write(example.SerializeToString())


# -*- coding: UTF-8 -*-
# Time: 8/15/17
# File: tftrainer.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals

import tensorflow as tf
import os
import re
import functools
from antgo.trainer.trainer import *
from antgo.trainer import tfmodel_deploy
from antgo.utils import logger
from antgo.measures.moving_statistic import *
from google.protobuf import text_format
from tensorflow.core.framework import graph_pb2
from tensorflow.python.platform import gfile
from antgo.utils.serialize import *
from antgo.utils.dht import *

import numpy as np
# from antgo.utils.net import *
slim = tf.contrib.slim


def _configure_optimizer(trianer_obj, learning_rate):
  """Configures the optimizer used for training.

  Args:
    learning_rate: A scalar or `Tensor` learning rate.

  Returns:
    An instance of an optimizer.

  Raises:
    ValueError: if optimizer is not recognized.
  """
  if trianer_obj.optimizer == 'adadelta':
    optimizer = tf.train.AdadeltaOptimizer(
        learning_rate,
        rho=trianer_obj.adadelta_rho,
        epsilon=trianer_obj.opt_epsilon)
  elif trianer_obj.optimizer == 'adagrad':
    optimizer = tf.train.AdagradOptimizer(
        learning_rate,
        initial_accumulator_value=trianer_obj.adagrad_initial_accumulator_value)
  elif trianer_obj.optimizer == 'adam':
    optimizer = tf.train.AdamOptimizer(
        learning_rate,
        beta1=trianer_obj.adam_beta1,
        beta2=trianer_obj.adam_beta2,
        epsilon=trianer_obj.opt_epsilon)
  elif trianer_obj.optimizer == 'ftrl':
    optimizer = tf.train.FtrlOptimizer(
        learning_rate,
        learning_rate_power=trianer_obj.ftrl_learning_rate_power,
        initial_accumulator_value=trianer_obj.ftrl_initial_accumulator_value,
        l1_regularization_strength=trianer_obj.ftrl_l1,
        l2_regularization_strength=trianer_obj.ftrl_l2)
  elif trianer_obj.optimizer == 'momentum':
    optimizer = tf.train.MomentumOptimizer(
        learning_rate,
        momentum=trianer_obj.momentum,
        name='Momentum')
  elif trianer_obj.optimizer == 'rmsprop':
    optimizer = tf.train.RMSPropOptimizer(
        learning_rate,
        decay=trianer_obj.rmsprop_decay,
        momentum=trianer_obj.rmsprop_momentum,
        epsilon=trianer_obj.opt_epsilon)
  elif trianer_obj.optimizer == 'sgd':
    optimizer = tf.train.GradientDescentOptimizer(learning_rate)
  else:
    raise ValueError('Optimizer [%s] was not recognized', trianer_obj.optimizer)
  return optimizer


def _configure_learning_rate(trainer_obj, num_samples_per_epoch, global_step):
  """Configures the learning rate.

  Args:
    num_samples_per_epoch: The number of samples in each epoch of training.
    global_step: The global_step tensor.

  Returns:
    A `Tensor` representing the learning rate.

  Raises:
    ValueError: if
  """
  batch_size = getattr(trainer_obj, 'batch_size', 1) * getattr(trainer_obj, 'num_clones', 1)
  num_epochs_per_decay = getattr(trainer_obj, 'num_epochs_per_decay', 10)
  decay_steps = int(num_samples_per_epoch / batch_size * num_epochs_per_decay)

  sync_replicas = getattr(trainer_obj, 'sync_replicas', False)
  if sync_replicas:
    replicas_to_aggregate = getattr(trainer_obj, 'replicas_to_aggregate', 1)
    decay_steps /= replicas_to_aggregate

  learning_rate_decay_type = getattr(trainer_obj, 'learning_rate_decay_type', 'exponential')
  learning_rate = getattr(trainer_obj, 'learning_rate', 0.1)
  if learning_rate_decay_type == 'exponential':
    learning_rate_decay_factor = getattr(trainer_obj, 'learning_rate_decay_factor', 0.9)
    return tf.train.exponential_decay(learning_rate,
                                      global_step,
                                      decay_steps,
                                      learning_rate_decay_factor,
                                      staircase=True,
                                      name='exponential_decay_learning_rate')
  elif learning_rate_decay_type == 'fixed':
    return tf.constant(learning_rate, name='fixed_learning_rate')
  elif learning_rate_decay_type == 'polynomial':
    end_learning_rate = getattr(trainer_obj, 'end_learning_rate', 0.0001)
    return tf.train.polynomial_decay(learning_rate,
                                     global_step,
                                     decay_steps,
                                     end_learning_rate,
                                     power=0.9,
                                     cycle=False,
                                     name='polynomial_decay_learning_rate')
  else:
    raise ValueError('learning_rate_decay_type [%s] was not recognized', learning_rate_decay_type)


def _get_variables_to_train(trainer_obj):
  """Returns a list of variables to train.

  Returns:
    A list of variables to train by the optimizer.
  """
  trainable_scopes = getattr(trainer_obj, 'trainable_scopes', None)
  if trainable_scopes is None:
    trainable_filter = getattr(trainer_obj, 'trainable_filter', None)
    if trainable_filter is None:
      return tf.trainable_variables()
    else:
      return [var for var in tf.trainable_variables() if var.name.startswith(trainable_filter)]
  else:
    scopes = [scope.strip() for scope in trainable_scopes.split(',')]

  variables_to_train = []
  for scope in scopes:
    variables = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope)
    variables_to_train.extend(variables)
  return variables_to_train


def _get_init_fn(trainer_obj, model, dump_dir, ctx=None):
  """Returns a function run by the chief worker to warm-start the training.

  Note that the init_fn is only run when initializing the model during the very
  first global step.

  Returns:
    An init function run by the supervisor.
  """
  # 1.step load from experiment
  if ctx is not None:
    if ctx.from_experiment is not None:
      no_activate_blocks = []
      for b in ctx.blocks():
        if not b.activate:
          no_activate_blocks.append(b.name)

      ablation_folder = '_'.join(sorted(no_activate_blocks))
      
      if len(ablation_folder) > 0:
        ctx.from_experiment = os.path.join(ctx.from_experiment, 'ablation', ablation_folder)
      # print(ctx.from_experiment)
      logger.info('load model from experiment %s' % ctx.from_experiment.split('/')[-2])
      latest_checkpoint = None
      try:
        latest_checkpoint = tf.train.latest_checkpoint(ctx.from_experiment)
      except:
        latest_checkpoint_index = -1
        for f in os.listdir(ctx.from_experiment):
          key_terms = f.split('.')
          if len(key_terms) >= 2 and key_terms[1] == 'ckpt':
            index = int(key_terms[0].split('_')[-1])
            if latest_checkpoint_index < index:
              latest_checkpoint_index = index
              latest_checkpoint = os.path.join(ctx.from_experiment, '.'.join(key_terms[0:2]))
      
      if latest_checkpoint is not None:
        variables_to_restore = {}
        # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
        model_variables = tf.global_variables()
        exclusions = []
        checkpoint_exclude_scopes = getattr(trainer_obj, 'checkpoint_exclude_scopes', None)
        if checkpoint_exclude_scopes is not None:
          exclusions = [scope.strip()
                        for scope in checkpoint_exclude_scopes.split(',')]
          
          logger.info('exclude scope %s from the latest checkpoint'%(','.join(exclusions)))

        fuzzy_exclusions = []
        checkpoint_fuzzy_exclude_scopes = getattr(trainer_obj, 'checkpoint_fuzzy_exclude_scopes', None)
        if checkpoint_fuzzy_exclude_scopes is not None:
          fuzzy_exclusions = [scope.strip()
                              for scope in checkpoint_fuzzy_exclude_scopes.split(',')]

          logger.info('exclude scope %s from the latest checkpoint' % (','.join(exclusions)))

        for var in model_variables:
          var_name = var.op.name

          excluded = False
          for exclusion in exclusions:
            if var_name.startswith(exclusion):
              excluded = True
              break

          for exclusion in fuzzy_exclusions:
            if exclusion in var_name:
              excluded = True
              break

          if excluded:
            continue
          
          variables_to_restore[var_name] = var
    
        return [slim.assign_from_checkpoint_fn(latest_checkpoint, variables_to_restore, ignore_missing_vars=True)]

  # Warn the user if a checkpoint exists in the train_dir. Then we'll be
  # ignoring the checkpoint anyway.
  # 2.step load from dump_dir
  if tf.train.latest_checkpoint(dump_dir):
    logger.info('Ignoring --checkpoint_path because a checkpoint already exists in %s'% dump_dir)
    # initilize model from dump_dir
    latest_checkpoint = tf.train.latest_checkpoint(dump_dir)
    variables_to_restore = {}
    # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
    model_variables = tf.global_variables()
    for var in model_variables:
      var_name = var.op.name
      variables_to_restore[var_name] = var

    logger.info('load model from %s' % latest_checkpoint)
    return [slim.assign_from_checkpoint_fn(latest_checkpoint, variables_to_restore)]
  
  # 3.step load from checkpoint_path
  checkpoint_path = getattr(trainer_obj, 'checkpoint_path', None)
  if checkpoint_path is None:
    return None

  exclusions = []
  checkpoint_exclude_scopes = getattr(trainer_obj, 'checkpoint_exclude_scopes', None)
  if checkpoint_exclude_scopes is not None:
    exclusions = [scope.strip()
                  for scope in checkpoint_exclude_scopes.split(',')]

  fuzzy_exclusions = []
  checkpoint_fuzzy_exclude_scopes = getattr(trainer_obj, 'checkpoint_fuzzy_exclude_scopes', None)
  if checkpoint_fuzzy_exclude_scopes is not None:
    fuzzy_exclusions = [scope.strip()
                  for scope in checkpoint_fuzzy_exclude_scopes.split(',')]

  transfers = {}
  checkpoint_transfer_scopes = getattr(trainer_obj, 'checkpoint_transfer_scopes', None)
  if checkpoint_transfer_scopes is not None:
    for scope in checkpoint_transfer_scopes.split(','):
      s_scope, t_scope = scope.split(':')
      transfers[s_scope.strip()] = t_scope.strip()
  
  # TODO(sguada) variables.filter_variables()
  auxilary_variables_to_restore = []
  variables_to_restore = {}
  # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
  model_variables = tf.global_variables()
  for var in model_variables:
    # transfer name
    var_name = var.op.name
    for k, v in transfers.items():
      if var.op.name.startswith(k):
        var_name = var_name.replace(k, v)
    
    # exclude name
    excluded = False
    for exclusion in exclusions:
      if var_name.startswith(exclusion):
        excluded = True
        break

    for exclusion in fuzzy_exclusions:
      if exclusion in var_name:
        excluded = True
        break

    # record
    if not excluded:
      if var_name not in variables_to_restore:
        variables_to_restore[var_name] = var
      else:
        is_ok = False
        for auxi_variables in auxilary_variables_to_restore:
          if var_name not in auxi_variables:
            auxi_variables[var_name] = var
            is_ok = True
        
        if not is_ok:
          auxilary_variables_to_restore.append({})
          auxilary_variables_to_restore[-1][var_name] = var
  
  auxilary_variables_to_restore.append(variables_to_restore)
  
  if tf.gfile.IsDirectory(checkpoint_path):
    checkpoint_path = tf.train.latest_checkpoint(checkpoint_path)
  
  logger.info('load model from %s' % checkpoint_path)
  return [slim.assign_from_checkpoint_fn(checkpoint_path, vr, ignore_missing_vars=True) for vr in auxilary_variables_to_restore if len(vr) > 0]


def _convert_to_svg_graph(tf_graph_pb_file, dump_dir, scopes):
  input_graph_def = graph_pb2.GraphDef()
  
  with gfile.FastGFile(tf_graph_pb_file, 'r') as f:
    text_format.Merge(f.read(), input_graph_def)
  
  my_graph = Graph(name='graph')
  # build group table (convolution and batchnorm)
  group_lookup_table = {}
  exclude_lookup_table = {}
  exclude_subfix = {}
  who_is_input = []
  for node in input_graph_def.node:
    terms = node.name.split('/')
    if node.op == 'Placeholder' or terms[0] == 'input':
      who_is_input.append(node.name)

    if terms[0] not in scopes:
      # check convolution group
      if terms[-1] == 'convolution' and node.op == 'Conv2D':
        # special flag
        group_lookup_table['/'.join(terms[0:-1])] = len(terms[0:-1])
      # check batchnorm group
      if len(terms) >= 2:
        if terms[-2] == 'batchnorm' and terms[-1] == 'Rsqrt':
          # special flag
          group_lookup_table['/'.join(terms[0:-2])] = len(terms[0:-2])
      # check AssignmovingAvg group
      if terms[-1].startswith('AssignMovingAvg'):
        group_lookup_table['/'.join(terms[0:-1])] = len(terms[0:-1])
      
      # check dropout group
      if 'dropout' in terms:
        p = 0
        for ti, t in enumerate(terms):
          if t == 'dropout':
            p = ti
        group_lookup_table['/'.join(terms[0:p + 1])] = p + 1
      
      # check optimizer
      if len(terms) >= 3:
        if 'update_discriminator' in terms:
          p = 0
          for ti, t in enumerate(terms):
            if t == 'update_discriminator':
              p = ti
          
          exclude_subfix[terms[p - 1]] = True
          exclude_lookup_table['/'.join(terms[0:p])] = p
      
      # check gradients group
      if len(terms) >= 2:
        if 'gradients' in terms:
          p = 0
          for ti, t in enumerate(terms):
            if t == 'gradients':
              p = ti
          exclude_lookup_table['/'.join(terms[0:p + 1])] = p + 1
      
      # check Relu in convolution scope
      if '/'.join(terms[0:-1]) in group_lookup_table:
        if terms[-1] == 'Relu':
          group_lookup_table['/'.join(terms)] = len(terms)
  
  node_name_map = {}
  for node in input_graph_def.node:
    terms = node.name.split('/')
    name = node.name
    if terms[0] not in scopes:
      # check const op
      if node.op == 'Const':
        continue
      
      # check special node
      if 'Initializer' in terms or 'Assign' in terms or 'read' in terms:
        continue
      
      # check exclude lookup table
      is_exclude = False
      for k, v in exclude_lookup_table.items():
        if name.startswith(k):
          is_exclude = True
          break
      if is_exclude:
        continue
      
      # check exclude subfix
      is_exclude = False
      for k, v in exclude_subfix.items():
        if terms[-1].startswith(k):
          is_exclude = True
          break
      if is_exclude:
        continue
      
      for k, v in group_lookup_table.items():
        if '/'.join(terms[0:v]) == k:
          # check larger match (+1)
          if '/'.join(terms[0:v + 1]) in group_lookup_table:
            name = '/'.join(terms[0:v + 1])
          else:
            name = '/'.join(terms[0:v])
          break
      
      node_name_map[node.name] = name
      # print('%s - %s' % (node.name, name))
      my_graph.add_node(name=name, op_name=name, label='l')
  
  for node in input_graph_def.node:
    terms = node.name.split('/')
    if terms[0] not in scopes:
      if node.op == 'Const':
        continue
      
      if 'Initializer' in terms or 'Assign' in terms or 'read' in terms:
        continue
      
      # check exclude lookup table
      is_exclude = False
      for k, v in exclude_lookup_table.items():
        if node.name.startswith(k):
          is_exclude = True
          break
      if is_exclude:
        continue
      
      # check exclude subfix
      is_exclude = False
      for k, v in exclude_subfix.items():
        if terms[-1].startswith(k):
          is_exclude = True
          break
      if is_exclude:
        continue
      
      input_name = node_name_map[node.name]
      if len(node.input) > 0:
        for linked_node in node.input:
          if linked_node not in node_name_map:
            continue
          
          linked_name = node_name_map[linked_node]
          if linked_name == input_name:
            continue
          
          my_graph.add_link(linked_name, input_name, Link(label='', args=''))
  
  # with open(os.path.join(dump_dir, 'graph.txt'), 'w') as fp:
  #   graph_str = Encoder().encode(my_graph)
  #   fp.write(graph_str)
  
  try:
    svg_graph = graph_net_visualization(my_graph, os.path.join(dump_dir, 'graph.svg'), who_is_input)
    return svg_graph
  except:
    logger.error('dont support graphviz')
    return None


class TFTrainer(Trainer):
  def __init__(self, dump_dir, is_training=True):
    super(TFTrainer, self).__init__(is_training)
    self.dump_dir = dump_dir

    self.saver = None
    self.sess = None
    self.graph = None
    self.clones = None
    self.lr = None

    self.coord = None
    self.threads = None

    self._model_dependence = None

    # record run time
    self.time_stat = MovingAverage(self.log_every_n_steps)
    self.loss_stat = MovingAverage(10)

    self._has_model_input = False
    self.cache = {}

    self.summary_op = None
    self.train_writer = None

    self._create_funcs = []
    self._aux_init_funcs = []

  def restore_scopy_from(self, model, restore_scope, checkpoint_path):
    # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
    model_variables = tf.global_variables()
    variables_to_restore = {}
    for var in model_variables:
      if var.op.name.startswith(restore_scope):
        variables_to_restore[var.op.name] = var

    if len(variables_to_restore) == 0:
      return

    if tf.gfile.IsDirectory(checkpoint_path):
      checkpoint_path = tf.train.latest_checkpoint(checkpoint_path)

    logger.info('restore %s scope from %s' % checkpoint_path)
    fn = slim.assign_from_checkpoint_fn(checkpoint_path, variables_to_restore, ignore_missing_vars=True)
    fn(self.sess)

  def run(self, *args, **kwargs):
    data_generator = None

    assert(len(args) <= 1)
    if len(args) != 0:
      data_generator = args[0]

    if data_generator is not None:
      return self._run_by_generator(data_generator, **kwargs)

    return self._run_by_feed(feed_dict=None, **kwargs)

  # 2.step run model once
  def _run_by_generator(self, data_generator, **kwargs):
    # bind data
    with self.graph.as_default():
      feed_dict = {}
      if self._has_model_input and len(self.clones) > 1:
        logger.error('clones number > 1, must set different placeholder for every clone')
        exit(-1)

      if self._has_model_input:
        # generate data
        data = next(data_generator)

        for k, v in kwargs.items():
          if k not in self.cache:
            placeholder_tensor = self.graph.get_tensor_by_name('{}/{}:0'.format('input', k))
            self.cache[k] = placeholder_tensor

          feed_dict[self.cache[k]] = data[v] if (type(data) == tuple or type(data) == list) else data
      else:
        # set different placeholder for every clone
        for clone in self.clones:
          # generate data
          data = next(data_generator)

          for k, v in kwargs.items():
            cache_name = '{}:0'.format(k)
            if len(self.clones) > 1:
              cache_name = '{}/{}:0'.format(clone[1][:-1], k)

            if cache_name not in self.cache:
              placeholder_tensor = None
              if len(self.clones) > 1:
                placeholder_tensor = self.graph.get_tensor_by_name('{}/{}:0'.format(clone[1][:-1], k))
              else:
                placeholder_tensor = self.graph.get_tensor_by_name('{}:0'.format(k))
              self.cache[cache_name] = placeholder_tensor

            feed_dict[self.cache[cache_name]] = data[v] if (type(data) == tuple or type(data) == list) else data

      return self._run_by_feed(feed_dict=feed_dict, **kwargs)

  def _run_by_feed(self, feed_dict=None, **kwargs):
    with self.graph.as_default():
      if feed_dict is None:
        feed_dict = {}
        for k_name, v_value in kwargs.items():
          if k_name not in self.cache:
            k_tensor = self.graph.get_tensor_by_name('{}:0'.format(k_name))
            self.cache[k_name] = k_tensor

          feed_dict[self.cache[k_name]] = v_value

      # forward process
      start_time = time.time()
      result = self.sess.run(self.val_ops, feed_dict=feed_dict if feed_dict is not None and len(feed_dict) > 0 else None)
      elapsed_time = int((time.time() - start_time) * 100) / 100.0

      if self.ctx.recorder is not None and self.ctx.recorder.model_fn is not None:
        self.ctx.recorder.action(result[-1])
        result = result[:-1]

      self.iter_at += 1

      if self.summary_op is not None:
        summary_op_val = result[0]
        result = result[1:]
        self.train_writer.add_summary(summary_op_val, self.iter_at)

      # record elapsed time
      self.time_stat.add(elapsed_time)

      # print log
      if self.is_training:
        loss_val = 0.0
        if type(result) == list and len(result) >= 2:
          loss_val = result[1]
        elif type(result) == list and len(result) == 1:
          loss_val = result[0]
        else:
          loss_val = result

        self.loss_stat.add(loss_val)
        if self.iter_at % self.log_every_n_steps == 0:
          if not self.is_distribute_training or(self.is_distribute_training and self.rank == 0):
            logger.info('(PID: %s) INFO: loss %f lr %f at iterator %d (%f sec/step)' %
                        (str(os.getpid()), self.loss_stat.get(), self.sess.run(self.lr), self.iter_at, float(self.time_stat.get())))
      else:
        if not self.is_distribute_training or (self.is_distribute_training and self.rank == 0):
         logger.info('(PID: %s) INFO: (%f sec/step)' % (str(os.getpid()), float(self.time_stat.get())))

      return result[0] if type(result) == list and len(result) == 1 else result

  # 3.step snapshot running state
  def snapshot(self, epoch=0, iter=-1):
    logger.info('snapshot at %d in %d epoch' % (self.iter_at, epoch))
    if not os.path.exists(self.dump_dir):
        os.makedirs(self.dump_dir)

    model_filename = "{prefix}_{infix}_{d}_{e}.ckpt".format(prefix=self.snapshot_prefix,
                                                        infix=self.snapshot_infix,
                                                        d=self.iter_at if iter < 0 else iter,
                                                        e=epoch)
    model_filepath = os.path.join(self.dump_dir, model_filename)

    # save checkpoint
    self.saver.save(self.sess, model_filepath)

  def training_deploy(self, model, funcs=[]):
    # Horovod: initialize Horovod (prepare MPI envoriment)
    if self.is_distribute_training:
      import horovod.tensorflow as hvd
      hvd.init()

      # reset num_clones = 1
      self.num_clones = 1
      self.rank = hvd.rank()
      self.local_rank = hvd.local_rank()

      get_global_context().quiet = False if self.rank == 0 else True

    tf.logging.set_verbosity(tf.logging.INFO)
    with tf.Graph().as_default() as graph:
      # Default graph
      self.graph = graph
      # Session
      config = tf.ConfigProto(allow_soft_placement=True)
      config.gpu_options.allow_growth = True
      devices = self.ctx.devices if len(self.ctx.devices) > 0 else self.devices
      config.gpu_options.visible_device_list = ','.join(str(x) for x in devices) if len(devices) > 0 else ''
      self.sess = tf.Session(graph=graph, config=config)
      
      #######################
      # Config model deploy #
      #######################
      deploy_config = tfmodel_deploy.DeploymentConfig(num_clones=self.num_clones,
                                                      devices=[],
                                                      clone_on_cpu=self.clone_on_cpu,
                                                      replica_id=self.replica_id,
                                                      num_replicas=self.worker_replicas,
                                                      num_ps_tasks=self.num_ps_tasks,
                                                      clone_id_map={0:self.local_rank} if self.is_distribute_training else {})

      # init some info
      with tf.device(deploy_config.inputs_device()):
        # Create global_step
        with tf.device(deploy_config.variables_device()):
          global_step = slim.get_or_create_global_step()

        ###################################
        ####    define model input (CPU) ##
        ###################################
        with tf.variable_scope('input'):
          data_queue = self.ctx.model.model_input(self.is_training)
          if data_queue is not None:
            self._has_model_input = True
        
        ###################################
        ####    define model (CPU or GPU) #
        ###################################
        func = model.model_fn
        @functools.wraps(func)
        def network_fn(*args, **kwargs):
          res = func(self.is_training, *args, **kwargs)
          if kwargs['clone'] == 0:
            # 1.step save graph file
            tf.train.write_graph(self.sess.graph_def, self.dump_dir, 'graph.pbtxt')
            
            # # 2.step transfer to local graph net
            # logger.info('build model graph svg')
            # svg_graph = _convert_to_svg_graph(os.path.join(self.dump_dir, 'graph.pbtxt'),
            #                                   self.dump_dir,
            #                                   ['input'])
            # if svg_graph is not None:
            #   self.ctx.job.send({'DATA': {'GRAPH': svg_graph}})
          return res

        ####################################
        ####### Create summary      ########
        ####################################
        summaries = set(tf.get_collection(tf.GraphKeys.SUMMARIES))

        ####################################
        ####### Create model clones ########
        ####################################
        self.clones = tfmodel_deploy.create_clones(deploy_config,
                                                   network_fn,
                                                   [data_queue] if data_queue is not None else None)
        first_clone_scope = deploy_config.clone_scope(0)
        update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS, first_clone_scope)

        for loss in tf.get_collection(tf.GraphKeys.LOSSES, first_clone_scope):
          summaries.add(tf.summary.scalar('losses/%s' % loss.op.name, loss))

        # create other func
        for create_func in funcs:
          self._create_funcs.append(create_func())

        #########################################
        # Configure the optimization procedure. #
        #########################################
        with tf.device(deploy_config.optimizer_device()):
          # samples total number
          num_samples = self.num_samples if self.num_samples > 0 else self.ctx.data_source.size

          # Horovod: adjust learning rate based on number of GPUs
          self.lr = _configure_learning_rate(self, num_samples, global_step)
          summaries.add(tf.summary.scalar('learning_rate', self.lr))

          # config optimizer
          optimizer = _configure_optimizer(self, self.lr)

          # Horovod: add Horovod Distributed Optimizer
          if self.is_distribute_training:
            optimizer = hvd.DistributedOptimizer(optimizer)

        # Variables to train.
        variables_to_train = _get_variables_to_train(self)
        
        with tf.control_dependencies(self.model_dependence):
          # Train_tensor
          total_loss, clones_gradients = \
            tfmodel_deploy.optimize_clones(self.clones,
                                           optimizer,
                                           regularization_losses=None if self.regularization_loss else [],
                                           var_list=variables_to_train)

          summaries.add(tf.summary.scalar('total_loss', total_loss))

          # Create gradient updates.
          grad_updates = optimizer.apply_gradients(clones_gradients,
                                                   global_step=global_step)

        # Value ops
        update_ops.append(grad_updates)
        update_op = tf.group(*update_ops)
        with tf.control_dependencies([update_op]):
          self.val_ops = tf.identity(total_loss, name='train_op')
  
          if self.clones[0].outputs is not None:
            self.val_ops = [self.val_ops]
            if type(self.clones[0].outputs) == list:
              self.val_ops.extend(self.clones[0].outputs)
            elif type(self.clones[0].outputs) == tuple:
              self.val_ops.extend(list(self.clones[0].outputs))
            else:
              self.val_ops.append(self.clones[0].outputs)

          if type(self.val_ops) != list:
            self.val_ops = [self.val_ops]

        summaries |= set(tf.get_collection(tf.GraphKeys.SUMMARIES,
                                           first_clone_scope))

        # Merge all summaries together.
        self.summary_op = tf.summary.merge(list(summaries), name='summary_op')

        if self.summary_op is not None:
          val_ops_temp = [self.summary_op]
          val_ops_temp.extend(self.val_ops)
          self.val_ops = val_ops_temp

        # summary write
        if not os.path.exists(os.path.join(self.dump_dir, 'summary')):
          os.makedirs(os.path.join(self.dump_dir, 'summary'))

        self.train_writer = tf.summary.FileWriter(os.path.join(self.dump_dir, 'summary'), graph)

        # Global initialization
        self.sess.run(tf.global_variables_initializer())
        self.sess.run(tf.local_variables_initializer())
        # coord
        self.coord = tf.train.Coordinator()
        self.threads = tf.train.start_queue_runners(sess=self.sess, coord=self.coord)
        
        custom_dataset_queue = tf.get_collection('CUSTOM_DATASET_QUEUE')
        if len(custom_dataset_queue) > 0:
          custom_dataset_queue[0].coord = self.coord
          custom_threads = custom_dataset_queue[0].start_threads(self.sess)
          self.threads.extend(custom_threads)
        
        # Training saver
        # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
        self.saver = tf.train.Saver(max_to_keep=2)

        # Restore from checkpoint
        if not self.is_distribute_training or (self.is_distribute_training and self.rank == 0):
          restore_fns = _get_init_fn(self, model, self.dump_dir, self.ctx)
          if restore_fns is not None:
            for restore_fn in restore_fns:
              restore_fn(self.sess)

          # Restore from custom auxilary init funcs
          for func in self._aux_init_funcs:
            func(self.sess)

        # resotre from auxilary checkpoint
        for auxilary_scope, auxilary_checkpoint in self.auxilary_checkpoints.items():
          self.restore_scopy_from(model, auxilary_scope, auxilary_checkpoint)

        # Horovod boardcast global variables
        if self.is_distribute_training:
          bgv = hvd.BroadcastGlobalVariablesHook(0)
          bgv.begin()
          bgv.after_create_session(self.sess, self.coord)

  def infer_deploy(self, model, funcs=[]):
    tf.logging.set_verbosity(tf.logging.INFO)
    with tf.Graph().as_default() as graph:
      # Default graph
      self.graph = graph
      # Session
      config = tf.ConfigProto(allow_soft_placement=True)
      config.gpu_options.allow_growth = True

      devices = self.ctx.devices if len(self.ctx.devices) > 0 else self.devices
      config.gpu_options.visible_device_list = ','.join(str(x) for x in devices) if len(devices) > 0 else ''

      self.sess = tf.Session(graph=graph, config=config)


      #######################
      # Config model_deploy #
      #######################
      deploy_config = tfmodel_deploy.DeploymentConfig(num_clones=1,
                                                      devices=[],
                                                      clone_on_cpu=getattr(self, 'clone_on_cpu', False),
                                                      replica_id=0,
                                                      num_replicas=1,
                                                      num_ps_tasks=0)
      
      # init some info
      with tf.device(deploy_config.inputs_device()):
        #############################
        ####    define model input ##
        #############################
        data_queue = None
        if self.ctx.ant is not None:
          with tf.variable_scope('input'):
            data_queue = self.ctx.model.model_input(self.is_training)
            if data_queue is not None:
              self._has_model_input = True

        #############################
        ####    define model       ##
        #############################
        func = model.model_fn
        @functools.wraps(func)
        def network_fn(*args, **kwargs):
          res = func(self.is_training, *args, **kwargs)
          return res
  
        #######################
        # Create model clones #
        #######################
        self.clones = tfmodel_deploy.create_clones(deploy_config, network_fn, [data_queue] if data_queue is not None else None)

        # create other func
        for create_func in funcs:
          self._create_funcs.append(create_func())

        # write graph
        tf.train.write_graph(graph.as_graph_def(), self.dump_dir, 'infer_graph.pbtxt')
        # svg_graph = _convert_to_svg_graph(os.path.join(self.dump_dir, 'infer_graph.pbtxt'),
        #                                   self.dump_dir,
        #                                   ['input'])
        # if svg_graph is not None:
        #   self.ctx.job.send({'DATA': {'GRAPH': svg_graph}})

        # Global initialization
        self.sess.run(tf.global_variables_initializer())
        self.sess.run(tf.local_variables_initializer())
        
        self.coord = tf.train.Coordinator()
        self.threads = tf.train.start_queue_runners(sess=self.sess, coord=self.coord)

        custom_dataset_queue = tf\
          .get_collection('CUSTOM_DATASET_QUEUE')
        if len(custom_dataset_queue) > 0:
          custom_dataset_queue[0].coord = self.coord
          custom_threads = custom_dataset_queue[0].start_threads(self.sess)
          self.threads.extend(custom_threads)

        # Restore from checkpoint
        restore_fns = _get_init_fn(self, model, self.dump_dir, self.ctx)
        if restore_fns is not None:
          for restore_fn in restore_fns:
            restore_fn(self.sess)

        # Restore from custom auxilary init funcs
        for func in self._aux_init_funcs:
          func(self.sess)

        # model saver
        # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
        self.saver = tf.train.Saver(max_to_keep=1)

        # snapshot
        self.snapshot(0)

        # Value ops
        self.val_ops = self.clones[0].outputs
        if type(self.val_ops) != list and type(self.val_ops) != tuple:
          self.val_ops = [self.val_ops]
        if type(self.val_ops) == tuple:
          self.val_ops = list(self.val_ops)

        # Append recorder model fn
        if self.ctx.recorder is not None and self.ctx.recorder.model_fn is not None:
          self.val_ops.append(self.ctx.recorder.model_fn)

  # 1.step deploy model on hardwares
  def deploy(self, model, funcs=[]):
    # model context
    self.ctx.model = model
    model.trainer = self

    # deploy model
    if self.is_training:
      self.training_deploy(model, funcs)
    else:
      self.infer_deploy(model, funcs)

      if self.ctx.ant is None and not self.ctx.debug:
        logger.info('successfully deploy model')
        exit(0)

  @property
  def deploy_funcs(self):
    return self._create_funcs

  @property
  def model_dependence(self):
    return self._model_dependence
  @model_dependence.setter
  def model_dependence(self, val):
    self._model_dependence = val
  
  def wait_until_clear(self):
    if self.coord is not None:
      self.coord.request_stop()
      self.coord.join(self.threads)

  def add_aux_init_func(self, func):
    self._aux_init_funcs.append(func)


# encoding=utf-8
# @Time    : 17-6-7
# @File    : regular.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
import cv2
from antgo.utils import get_rng
from antgo.dataflow.core import *
from scipy.ndimage.interpolation import affine_transform
import copy

class Flip(Node):
  def __init__(self, inputs=None, horiz=True, vert=False, prob=0.5):
    super(Flip, self).__init__(name=None, action=self.action, inputs=inputs)

    """
    Only one of horiz, vert can be set.

    :param horiz: whether or not apply horizontal flip.
    :param vert: whether or not apply vertical flip.
    :param prob: probability of flip.
    """
    if horiz and vert:
      raise ValueError("Please use two Flip instead.")
    elif horiz:
      self.code = 1
    elif vert:
      self.code = 0
    else:
      raise ValueError("Are you kidding?")
    self.prob = prob
    self.rng = get_rng(self)

  def action(self, *args, **kwargs):
    assert(len(args) == 1)
    is_flip = self.rng.uniform(0, 1) < self.prob
    is_single = True
    if type(args[0]) == tuple or type(args[0]) == list:
      is_single = False
    image, annotation = args[0] if type(args[0]) == tuple or type(args[0]) == list else (args[0], {})
    
    annotation = copy.deepcopy(annotation)
    if is_flip:
      # for image
      if self.code == 1:
        image = np.fliplr(image)
      else:
        image = np.flipud(image)

      # for annotation
      if 'bbox' in annotation:
        if self.code == 1:
          w = annotation['info'][1]
          boxes = annotation['bbox'].copy()
          oldx1 = boxes[:, 0].copy()
          oldx2 = boxes[:, 2].copy()
          boxes[:, 0] = w - oldx2 - 1
          boxes[:, 2] = w - oldx1 - 1

          assert ((boxes[:, 2] >= boxes[:, 0]).all())
          annotation['bbox'] = boxes
          annotation['flipped'] = True
        else:
          h = annotation['info'][0]
          boxes = annotation['bbox'].copy()
          oldy1 = boxes[:, 1].copy()
          oldy2 = boxes[:, 3].copy()
          boxes[:, 1] = h - oldy2 - 1
          boxes[:, 3] = h - oldy1 - 1

          assert ((boxes[:, 3] >= boxes[:, 1]).all())
          annotation['bbox'] = boxes
          annotation['flipped'] = True
      if 'segmentation' in annotation:
        resized_obj_seg = []
        for obj_seg in annotation['segmentation']:
          temp = None
          if self.code == 1:
            temp = np.fliplr(obj_seg)
          else:
            temp = np.flipud(obj_seg)
          resized_obj_seg.append(temp)
        annotation['segmentation'] = resized_obj_seg

    return (image, annotation) if not is_single else image


class Resize(Node):
  def __init__(self, inputs=None, shape=None):
    super(Resize, self).__init__(name=None, action=self.action, inputs=inputs)
    self.shape = shape

  def action(self, *args, **kwargs):
    assert (len(args) == 1)
    is_single = True
    if type(args[0]) == tuple or type(args[0]) == list:
      is_single = False
      
    image, annotation = args[0] if type(args[0]) == tuple or type(args[0]) == list else (args[0], {})
    # for image
    # image = scipy.misc.imresize(image, self.shape)
    image = cv2.resize(image, self.shape)
    annotation = copy.deepcopy(annotation)
    # for annotation
    if type(annotation) == dict and 'bbox' in annotation:
      boxes = annotation['bbox']
      info = annotation['info']
      horizontal_scale = float(self.shape[1]) / float(info[1])
      vertical_scale = float(self.shape[0]) / float(info[0])

      boxes[:, [0, 2]] = boxes[:, [0, 2]] * horizontal_scale
      boxes[:, [1, 3]] = boxes[:, [1, 3]] * vertical_scale

      annotation['bbox'] = boxes
    if type(annotation) == dict and 'segmentation' in annotation:
      resized_obj_seg = []
      for obj_seg in annotation['segmentation']:
        # temp = scipy.misc.imresize(obj_seg[:, :, 0], self.shape[::-1], 'nearest')
        temp = cv2.resize(obj_seg[:, :, 0], self.shape[::-1], cv2.INTER_NEAREST)
        resized_obj_seg.append(temp.reshape(temp.shape[0],temp.shape[1],1))
      annotation['segmentation'] = resized_obj_seg

    return (image, annotation) if not is_single else image


class Subtract(Node):
  def __init__(self, inputs=None, mean_val=(128, 128, 128)):
    super(Subtract, self).__init__(name=None, action=self.action, inputs=inputs)
    self._mean_value = mean_val

  def action(self, *args, **kwargs):
    assert(len(args) == 1)
    is_single = True
    if type(args[0]) == tuple or type(args[0]) == list:
      is_single = False

    image, annotation = args[0] if type(args[0]) == tuple or type(args[0]) == list else (args[0], {})

    # for image
    image = image - np.array(self._mean_value)

    return (image, annotation) if not is_single else image


class DisturbChannels(Node):
  def __init__(self, inputs=None):
    super(DisturbChannels, self).__init__(name=None, action=self.action, inputs=inputs)

  def action(self, *args, **kwargs):
    assert(len(args) == 1)
    is_single = True
    if type(args[0]) == tuple or type(args[0]) == list:
      is_single = False

    image, annotation = args[0] if type(args[0]) == tuple or type(args[0]) == list else (args[0], {})
    
    if len(image.shape) < 3:
      return [image, annotation]

    channles = np.arange(0, image.shape[2])
    np.random.shuffle(channles)
    image_pieces = np.split(image, len(channles), axis=2)
    image = np.concatenate([image_pieces[i] for i in channles], axis=2)

    return (image, annotation) if not is_single else image


class DisturbRotation(Node):
  def __init__(self, inputs=None, max_deg=10):
    super(DisturbRotation, self).__init__(name=None, action=self.action, inputs=inputs)
    self._max_deg = max_deg
    self.rng = get_rng(self)

  @staticmethod
  def _affine_matrix(rotate_center, rotate_deg):
    theta = np.deg2rad(rotate_deg)
    rot = np.array([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]])
    # offset = np.array(rotate_center) - np.array(rotate_center).dot(rot)
    offset = (rotate_center - rotate_center.dot(rot)).dot(np.linalg.inv(rot))

    return rot, -offset

  def action(self, *args, **kwargs):
    assert(len(args) == 1)
    is_single = True
    if type(args[0]) == tuple or type(args[0]) == list:
      is_single = False

    image, annotation = args[0] if type(args[0]) == tuple or type(args[0]) == list else (args[0], {})

    deg = self.rng.uniform(-self._max_deg, self._max_deg)
    center_x = image.shape[1] / 2.0
    center_y = image.shape[0] / 2.0
    rot, offset = DisturbRotation._affine_matrix(np.array([center_y, center_x]), -deg)

    image_channles = [None, None, None]
    for channel, image_channle in enumerate(np.split(image, 3, axis=2)):
      image_channles[channel] = \
          affine_transform(np.squeeze(image_channle, 2),
                           rot,
                           order=3,
                           offset=offset,
                           cval=0,
                           output=np.uint8)

    image = np.concatenate((np.expand_dims(image_channles[0], 2),
                            np.expand_dims(image_channles[1], 2),
                            np.expand_dims(image_channles[2], 2)), axis=2)
    
    annotation_cpy = copy.deepcopy(annotation)
    if type(annotation_cpy) == dict and 'bbox' in annotation_cpy:
      bbox = annotation_cpy['bbox']

      rrr_theta = np.deg2rad(deg)
      rrr = np.array([[np.cos(rrr_theta), np.sin(rrr_theta)], [-np.sin(rrr_theta), np.cos(rrr_theta)]])

      # x0,y0,x1,y1,x0,y1,x1,y0
      bbox_mm = np.concatenate((bbox,
                                bbox[:, 0][:, np.newaxis], bbox[:, 3][:, np.newaxis],
                                bbox[:, 2][:, np.newaxis], bbox[:, 1][:, np.newaxis]), axis=1)
      bbox_mm = bbox_mm.reshape(-1, 2)
      rotate_bbox = np.matmul(bbox_mm - np.array([center_x, center_y]).reshape(-1, 2), rrr) + \
                    np.array([center_x, center_y]).reshape(-1, 2)
      rotate_bbox = rotate_bbox.reshape(-1, 8)

      # reset bbox and clip
      temp = np.zeros(bbox.shape)
      for bbox_i, bbox in enumerate(rotate_bbox):
        x0, y0, x1, y1, xx0, yy0, xx1, yy1 = bbox
        n_x0 = np.minimum(np.maximum(np.min([x0, x1, xx0, xx1]), 0), image.shape[1] - 1)
        n_y0 = np.minimum(np.maximum(np.min([y0, y1, yy0, yy1]), 0), image.shape[0] - 1)

        n_x1 = np.maximum(np.minimum(np.max([x0, x1, xx0, xx1]), image.shape[1] - 1), 0)
        n_y1 = np.maximum(np.minimum(np.max([y0, y1, yy0, yy1]), image.shape[0] - 1), 0)

        temp[bbox_i, 0] = n_x0
        temp[bbox_i, 1] = n_y0
        temp[bbox_i, 2] = n_x1
        temp[bbox_i, 3] = n_y1

      annotation_cpy['bbox'] = temp

    if type(annotation_cpy) == dict and 'segmentation' in annotation_cpy:
      obj_segs = []
      for obj_seg in annotation_cpy['segmentation']:
        obj_segs.append(
          np.expand_dims(affine_transform(np.squeeze(obj_seg, 2),
                                          rot,
                                          order=3,
                                          offset=offset,
                                          cval=0,
                                          output=np.uint8), axis=2))

      annotation_cpy['segmentation'] = obj_segs
    return (image, annotation_cpy) if not is_single else image


class DisturbLighting(Node):
  def __init__(self, inputs=None, max_lighting=10):
    super(DisturbLighting, self).__init__(name=None, action=self.action, inputs=inputs)
    self._max_lighting = max_lighting
    self.rng = get_rng(self)

  def action(self, *args, **kwargs):
    assert(len(args) == 1)
    is_single = True
    if type(args[0]) == tuple or type(args[0]) == list:
      is_single = False

    image, annotation = args[0] if type(args[0]) == tuple or type(args[0]) == list else (args[0], {})
    
    light = self.rng.uniform(-self._max_lighting, self._max_lighting)
    image_cpy = image.copy() + light

    return (image_cpy, annotation) if not is_single else image_cpy


class DisturbNoise(Node):
  def __init__(self, inputs=None, sigma=10):
    super(DisturbNoise, self).__init__(name=None, action=self.action, inputs=inputs)
    self._sigma = sigma
    self.rng = get_rng(self)

  def action(self, *args, **kwargs):
    assert(len(args) == 1)
    is_single = True
    if type(args[0]) == tuple or type(args[0]) == list:
      is_single = False

    image, annotation = args[0] if type(args[0]) == tuple or type(args[0]) == list else (args[0], {})
    
    height = image.shape[0]
    width = image.shape[1]
    channels = image.shape[2]
    noise = (np.random.random((height, width, channels)) * 2 - 1) * self._sigma
    image_cpy = image.copy() + noise

    return (image_cpy, annotation) if not is_single else image_cpy



# -*- coding: UTF-8 -*-
# @Time    : 17-9-5
# @File    : generate.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.ant.base import *
from antgo.dataflow.basic import *
from antgo.utils.dht import *
from antgo.utils.serialize import *

if sys.version > '3':
    PY3 = True
else:
    PY3 = False

Config = config.AntConfig


class AntGenerate(AntBase):
  def __init__(self, ant_context,
               ant_name,
               ant_data_folder,
               dump_dir,
               ant_token,
               dataset_name):
    super(AntGenerate, self).__init__(ant_name, ant_context, ant_token)
    self.ant_data_folder = ant_data_folder
    self.ant_context.ant = self
    self.dataset_name = dataset_name
    self.dump_dir = dump_dir

  def start(self):
    # 1.step check basic parameters
    if self.context.data_generator is None:
      logger.error('Must set data generator.')
      return

    if self.dataset_name is None:
      logger.error('Must set dsataset name.')
      return

    # 2.step build experiment folder
    now_time_stamp = datetime.fromtimestamp(self.time_stamp).strftime('%Y%m%d.%H%M%S.%f')
    if not os.path.exists(os.path.join(self.dump_dir, now_time_stamp, self.dataset_name)):
      os.makedirs(os.path.join(self.dump_dir, now_time_stamp, self.dataset_name))
    experiment_folder = os.path.join(self.dump_dir, now_time_stamp, self.dataset_name)
    logger.info('Dataset would be saved in %s.'%now_time_stamp)

    # 3.step build dataset (rocksdb)
    # dataset "train", 'val', 'test'
    for dataset_flag in ['train', 'val', 'test']:
      dataset_generateor = self.context.data_generator(dataset_flag)
      if dataset_generateor is not None:
        logger.info('Start build %s set for %s.'%(dataset_flag, self.dataset_name))
        if not os.path.exists(os.path.join(experiment_folder, dataset_flag)):
          os.makedirs(os.path.join(experiment_folder, dataset_flag))

        record_writer = RecordWriter(os.path.join(experiment_folder, dataset_flag))
        try:
          for data, label in dataset_generateor:
            record_writer.write(Sample(data=data, label=label))
        except:
          pass

        logger.info('Finish build %s set for %s.'%(dataset_flag, self.dataset_name))
      else:
        logger.warn('Couldn build %s set for %s.'%(dataset_flag, self.dataset_name))



# encoding=utf-8
# @Time    : 17-7-25
# @File    : flags.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
import getopt
_options = {}


class _FLAGS(object):
  def __init__(self):
    self._flags = []
    self._flags_name = []

  def add(self, flag):
    setattr(self, flag.name, lambda: flag.value)
    self._flags.append(flag)
    self._flags_name.append(flag.name)

  @property
  def shortopts(self):
    return ':'.join(self._flags_name)

  @property
  def longopts(self):
    a =  [f.name + "=" for f in self._flags if f.has_value]
    a.extend([f.name for f in self._flags if not f.has_value])
    return a

AntFLAGS = _FLAGS()


class VarFlag(object):
  def __init__(self, name, default, var_help, has_value=True):
    self._name = name
    self._var_help = var_help
    self._default = default
    self._has_value = has_value

  @property
  def name(self):
    return self._name
  @property
  def help(self):
    return self._var_help
  @property
  def default(self):
    return self._default
  @property
  def has_value(self):
    return self._has_value

  @property
  def value(self):
    raise NotImplementedError


class _FloatFlag(VarFlag):
  def __init__(self, name, default, var_help):
    super(_FloatFlag, self).__init__(name, default, var_help)

  @property
  def value(self):
    global _options
    if '--' + self.name in _options:
      return float(_options['--' + self.name])
    elif '-' + self.name in _options:
      return float(_options['--' + self.name])

    return float(self.default) if self.default is not None else None


def DEFINE_float(name, default, var_help):
  global AntFLAGS
  AntFLAGS.add(_FloatFlag(name, default, var_help))


class _IntegerFlag(VarFlag):
  def __init__(self, name, default, var_help):
    super(_IntegerFlag, self).__init__(name, default, var_help)

  @property
  def value(self):
    global _options
    if '--' + self.name in _options:
      return int(_options['--' + self.name])
    elif '-' + self.name in _options:
      return int(_options['--' + self.name])

    return int(self.default) if self.default is not None else None


def DEFINE_integer(name, default, var_help):
  global AntFLAGS
  AntFLAGS.add(_IntegerFlag(name, default, var_help))


class _StringFlag(VarFlag):
  def __init__(self, name, default, var_help):
    super(_StringFlag, self).__init__(name, default, var_help)

  @property
  def value(self):
    global _options
    if '--' + self.name in _options:
      return str(_options['--' + self.name])
    elif '-' + self.name in _options:
      return str(_options['--' + self.name])

    return str(self.default) if self.default is not None else None


def DEFINE_string(name, default, var_help):
  global AntFLAGS
  AntFLAGS.add(_StringFlag(name, default, var_help))


class _BooleanFlag(VarFlag):
  def __init__(self, name, default, var_help):
    super(_BooleanFlag, self).__init__(name, default, var_help)

  @property
  def value(self):
    global _options
    if '--' + self.name in _options:
      return bool(int(_options['--' + self.name]))
    elif '-' + self.name in _options:
      return bool(int(_options['--' + self.name]))

    return bool(self.default) if self.default is not None else None


def DEFINE_boolean(name, default, var_help):
  global AntFLAGS
  AntFLAGS.add(_BooleanFlag(name, default, var_help))


class _IndicatorFlag(VarFlag):
  def __init__(self, name, var_help):
    super(_IndicatorFlag, self).__init__(name, '', var_help, False)

  @property
  def value(self):
    global _options
    if '--' + self.name in _options:
      return True
    elif '-' + self.name in _options:
      return True

    return False

def DEFINE_indicator(name, var_help):
  global AntFLAGS
  AntFLAGS.add(_IndicatorFlag(name, var_help))


def cli_param_flags(cmd_str=None):
  global _options
  global AntFLAGS

  if cmd_str is None:
    cmd_str = sys.argv[2:]

  kv_args, args = getopt.getopt(cmd_str, AntFLAGS.shortopts, AntFLAGS.longopts)
  for k, v in kv_args:
    _options[k] = v

def clear_cli_param_flags():
  global _options
  _options = {}


from __future__ import unicode_literals
from __future__ import division

import numpy as np
import unittest


def pr(actual,predicated_s):
    '''
    computing precision recall curve
    what's difference between ROC and precision-recall curve?
    ROC curve plot True Positive Rate Vs. False Positive Rate;
    Whereas, PR curve plot Precision Vs. Recall.Particularly,
    if true negative is not much valuable to the problem,
    or negative examples are abundant.
    Then, PR-curve is typically more appropriate.
    For example, if the class is highly imbalanced and
    positive samples are very rare, then use PR-curve.
    One example may be fraud detection, where non-fraud
    sample may be 10000 and fraud sample may be below 100.
    In other cases, ROC curve will be more helpful.
    (only for binary classificaiton or information retrieval task)
    Parameters
    ----------
    actual: list
            ground truth label [1,0,0,1.....]
    predicated_s: list
            predicated score [0.5,0.2,....]

    Returns
    -------
    recall(x): recall list
    precision(y): precision list
    '''
    if type(actual) == list:
        actual = np.array(actual)
    if type(predicated_s) == list:
        predicated_s = np.array(predicated_s)

    if actual.shape[0] == 0 or predicated_s.shape[0] == 0:
        return None

    # transform to array(int)
    if actual.dtype != int:
        actual = actual.astype(dtype=int)

    # absorb some error
    if np.max(actual) > 1:
        actual[np.where(actual > 1)[0]] = 1
    if len(predicated_s.shape) > 1 and predicated_s.shape[1] > 1:
        predicated_s = predicated_s[:,0]
    if len(predicated_s.shape) > 1:
        predicated_s = predicated_s.flatten()

    sorted_x = sorted(zip(predicated_s, range(predicated_s.shape[0])), reverse=True)

    p_num = len(np.where(actual == 1)[0])
    precision = []
    recall = []
    num_hits = 0.0
    re_thres = np.array(range(11)) * 0.1

    for i in range(len(sorted_x)):
        if predicated_s[sorted_x[i][1]] <= -float("inf"):
            continue

        if actual[sorted_x[i][1]] == 1:
            num_hits += 1.0

            cur_recall = num_hits / float(p_num)
            cur_precision = num_hits / (i + 1.0)

            recall.append(cur_recall)
            precision.append(cur_precision)

    inds = np.searchsorted(recall, re_thres, side='left')
    pr = [-1 for _ in range(len(inds))]
    rc = [-1 for _ in range(len(inds))]
    try:
        for ri, pi in enumerate(inds):
            if ri > 0 and pi == inds[ri - 1]:
                continue

            if pi == len(precision):
                pr[ri] = precision[pi - 1]
                rc[ri] = recall[pi - 1]
            else:
                pr[ri] = precision[pi]
                rc[ri] = recall[pi]
    except:
        pass

    pr = [_ for _ in pr if _ > 0]
    rc = [_ for _ in rc if _ > 0]

    pr = np.array([rc, pr])
    return pr.transpose()


def pr_f1(actual,predicated_s, k = 10):
    '''
    summarizing the precision-recall curve with fixed k
    Parameters
    ----------
    actual: list
            ground truth label
    predicated_s: list
            predicated score
    k: int
            The maximum number of predicted elements

    Returns
    -------
    F1 value
    '''
    if type(actual) == list:
        actual = np.array(actual)
    if type(predicated_s) == list:
        predicated_s = np.array(predicated_s)

    if actual.shape[0] == 0 or predicated_s.shape[0] == 0:
        return None

    # transform to array(int)
    if actual.dtype != int:
        actual = actual.astype(dtype=int)

    # absorb some error
    if np.max(actual) > 1:
        actual[np.where(actual > 1)[0]] = 1
    if len(predicated_s.shape) > 1 and predicated_s.shape[1] > 1:
        predicated_s = predicated_s[:,0]
    if len(predicated_s.shape) > 1:
        predicated_s = predicated_s.flatten()

    sorted_x = sorted(zip(predicated_s, range(predicated_s.shape[0])), reverse=True)

    p_num = len(np.where(actual == 1)[0])
    top_k = min(k, len(sorted_x))

    num_hits = 0.0
    for i in range(top_k):
        if actual[sorted_x[i][1]] == 1:
            num_hits += 1.0

    precision = num_hits / (k + 0.000001)
    recall = num_hits / (p_num + 0.000001)
    F1 = 2.0 * (precision * recall) / (precision + recall + 0.000001)
    return F1

def dcg_k(r,k,method = 0):
    '''
    Score is discounted cumulative gain (dcg)
    Relevance is positive real values.  Can use binary
    as the previous methods.
    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf
    Parameters
    ----------
    r: Relevance scores (list or numpy) in rank order
            (first element is the first item)
    k: Number of results to consider
    method: 0 or 1

    Returns
    -------
    Discounted cumulative gain
    '''
    r = np.asfarray(r)[:k]
    if r.size:
        if method == 0:
            #standard definition
            return r[0] + np.sum(r[1:] / np.log2(np.arange(2, r.size + 1)))
        elif method == 1:
            #used in Kaggle
            return np.sum((np.power(2,r) - 1.0) / np.log2(np.arange(2, r.size + 2)))
        else:
            raise ValueError('method must be 0 or 1.')
    return 0.


def ndcg_k(r,k,method=0):
    '''
    Score is normalized discounted cumulative gain (ndcg)
    Relevance is positive real values.  Can use binary
    as the previous methods.
    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf
    Parameters
    ----------
    r: Relevance scores (list or numpy) in rank order
            (first element is the first item)
    k: Number of results to consider
    method: 0 or 1

    Returns
    -------
    Normalized discounted cumulative gain
    '''
    dcg_max = dcg_k(sorted(r, reverse=True), k, method)
    if not dcg_max:
        return 0.
    return dcg_k(r, k, method) / dcg_max

class TestPR(unittest.TestCase):
    def test_func(self):
        actual = [0, 1, 1, 1]
        predicated_s = [0.3, 0.1, 0.2, 0.4]

        recall, precision = pr(actual, predicated_s)
        np.testing.assert_array_almost_equal(recall, [0.3333333333333333, 0.6666666666666666, 1.0])
        np.testing.assert_array_almost_equal(precision, [1.0, 0.6666666666666666, 0.75])

        f1_v = pr_f1(actual, predicated_s, k=2)
        self.assertAlmostEqual(f1_v, 2.0 * ((1.0 / 2.0) * (1.0 / 3.0)) / (1.0 / 2.0 + 1.0 / 3.0))

        r = [3, 2, 3, 0, 0, 1, 2, 2, 3, 0]
        # dcg test
        self.assertAlmostEqual(dcg_k(r, 1), 3.0)
        self.assertAlmostEqual(dcg_k(r, 2), 5.0)

        # ndcg test
        r = [3, 2, 3, 0, 0, 1, 2, 2, 3, 0]
        self.assertAlmostEqual(ndcg_k(r, 4), 0.77509868485972222)

if __name__ == "__main__":
    unittest.main()


#encoding=utf-8
from __future__ import unicode_literals
from __future__ import division

import numpy as np

def cross_entropy(actual,predicated_score,confidence_interval=0.95):
    '''
    compute cross entropy (for multi-classification task and binary-classification task)
    Parameters
    ----------
    actual: list or ndarray
            ground truth label
    predicated_s: list or ndarray
            predicated score for every label

    Returns
    -------

    '''
    # type consistent
    if type(actual) == list:
        actual = np.array(actual, dtype=np.int32)
    if type(predicated_score) == list:
        predicated_score = np.array(predicated_score)

    if actual.dtype != np.int32:
        actual = actual.astype(dtype=np.int32)

    # shape consistent
    assert (len(actual.shape) == 1)
    assert (len(predicated_score.shape) == 2)
    assert (actual.shape[0] == predicated_score.shape[0])

    # class number
    C_N = predicated_score.shape[1]
    assert(C_N >= 2)
    # sample number
    S_N = predicated_score.shape[0]

    # normalize predicated score
    ss = np.sum(predicated_score,axis=1)
    predicated_s = np.reshape(predicated_score,[S_N,1,C_N]) / np.reshape(ss,[S_N,1,1])
    predicated_s = np.reshape(predicated_s,[S_N,C_N])
    loss = 0.0

    # cross entropy
    for n in range(C_N):
        index = np.where(actual == n)[0]
        loss += sum(-np.log(predicated_s[index,n]))

    return loss / float(S_N)

def multi_accuracy(actual,predicated_score):
    '''
    compute classification accuracy(for multi-classification task and binary-classification task)
    Parameters
    ----------
    actual: list or ndarray
            ground truth list
    predicated_score: list or ndarray
            predicated score for every label

    Returns
    -------
    '''
    # type consistent
    if type(actual) == list:
        actual = np.array(actual,dtype=np.int32)
    if type(predicated_score) == list:
        predicated_score = np.array(predicated_score)

    if actual.shape[0] == 0 or predicated_score.shape[0] == 0:
        return None

    if actual.dtype != np.int32:
        actual = actual.astype(dtype=np.int32)

    # shape consistent
    assert(len(actual.shape) == 1)
    assert(len(predicated_score.shape) == 2)
    assert(actual.shape[0] == predicated_score.shape[0])

    #predicated label
    max_index = np.argmax(predicated_score,axis=1)
    index = np.where(max_index == actual)
    return len(index[0]) / float(actual.shape[0])


def multi_accuracy_labels(actual, predicated_label):
    '''
    compute classification accuracy(for multi-classification task and binary-classification task)
    Parameters
    ----------
    actual: list or ndarray
            ground truth list
    predicated_score: list or ndarray
            predicted label list
    Returns
    -------
    '''
    # type consistent
    if type(actual) == list:
        actual = np.array(actual, dtype=np.int32)
    if type(predicated_label) == list:
        predicated_label = np.array(predicated_label)

    if actual.shape[0] == 0 or predicated_label.shape[0] == 0:
        return None

    if actual.dtype != np.int32:
        actual = actual.astype(dtype=np.int32)

    # shape consistent
    assert (len(actual.shape) == 1)
    assert (len(predicated_label.shape) == 1)

    # predicated label
    index = np.where(predicated_label == actual)
    return len(index[0]) / float(actual.shape[0])

def per_class_accuracy(actual,predicated_score):
    '''
    compute mean per-class classification accuracy(for multi-classification and binary-classification task)
    It's just for imbalance data
    Parameters
    ----------
    actual: list or ndarray
            ground truth label
    preicated_score: list or ndarray
            predicated score for every label
    Returns
    -------
    return mean classification accuracy
    '''
    # type consistent
    if type(actual) == list:
        actual = np.array(actual, dtype=np.int32)
    if type(predicated_score) == list:
        predicated_score = np.array(predicated_score)

    if actual.shape[0] == 0 or predicated_score.shape[0] == 0:
        return None

    if actual.dtype != np.int32:
        actual = actual.astype(dtype=np.int32)

    # shape consistent
    assert (len(actual.shape) == 1)
    assert (len(predicated_score.shape) == 2)
    assert (actual.shape[0] == predicated_score.shape[0])

    # class number
    c_num = predicated_score.shape[1]

    max_arg = np.argmax(predicated_score,axis=1)
    per_class_accuracy_array = np.ones(c_num)
    for c_index in range(c_num):
        index = np.where(actual == c_index)[0]
        predicated_label = max_arg[index]
        per_class_accuracy_array[c_index] = len(np.where(predicated_label == c_index)[0]) / float(len(index))

    return per_class_accuracy_array.tolist()


# -*- coding: UTF-8 -*-
# @Time    : 17-12-14
# @File    : preprocess.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import tensorflow as tf
from tensorflow.python.ops import control_flow_ops
import tensorflow.contrib.slim as slim


def tf_regular_augumentation(image,
                             brightness_max_delta=32./255.,
                             saturation_lower=0.5,
                             saturation_upper=1.5,
                             hue_max_delta=0.2,
                             contrast_lower=0.5,
                             contrast_upper=1.5):
  image = tf.image.random_brightness(image, max_delta=brightness_max_delta)
  image = tf.image.random_saturation(image, lower=saturation_lower, upper=saturation_upper)
  image = tf.image.random_hue(image, max_delta=hue_max_delta)
  image = tf.image.random_contrast(image, lower=contrast_lower, upper=contrast_upper)

  #image = tf.clip_by_value(image, 0, 255)
  return image


def _crop(image, offset_height, offset_width, crop_height, crop_width):
  original_shape = tf.shape(image)

  rank_assertion = tf.Assert(
      tf.equal(tf.rank(image), 3),
      ['Rank of image must be equal to 3.'])
  cropped_shape = control_flow_ops.with_dependencies(
      [rank_assertion],
      tf.stack([crop_height, crop_width, original_shape[2]]))

  size_assertion = tf.Assert(
      tf.logical_and(
          tf.greater_equal(original_shape[0], crop_height),
          tf.greater_equal(original_shape[1], crop_width)),
      ['Crop size greater than the image size.'])

  offsets = tf.to_int32(tf.stack([offset_height, offset_width, 0]))

  # Use tf.slice instead of crop_to_bounding box as it accepts tensors to
  # define the crop size.
  image = control_flow_ops.with_dependencies([size_assertion], tf.slice(image, offsets, cropped_shape))
  return tf.reshape(image, cropped_shape)


def tf_random_crop(image, label, crop_height, crop_width):
  image_rank = tf.rank(image)
  rank_assert = tf.Assert(tf.equal(image_rank, 3),
    ['Wrong rank for tensor  %s [expected] [actual]', image.name, 3, image_rank])
    
  image_shape = control_flow_ops.with_dependencies([rank_assert], tf.shape(image))
  image_height = image_shape[0]
  image_width = image_shape[1]
  crop_size_assert = tf.Assert(
      tf.logical_and(
          tf.greater_equal(image_height, crop_height),
          tf.greater_equal(image_width, crop_width)),
      ['Crop size greater than the image size.', image_height, image_width, crop_height, crop_width])

  asserts = [rank_assert, crop_size_assert]

  shape = control_flow_ops.with_dependencies([rank_assert], tf.shape(image))
  height = shape[0]
  width = shape[1]

  height_assert = tf.Assert(
      tf.equal(height, image_height),
      ['Wrong height for tensor %s [expected][actual]',
       image.name, height, image_height])
  width_assert = tf.Assert(
      tf.equal(width, image_width),
      ['Wrong width for tensor %s [expected][actual]',
       image.name, width, image_width])
  asserts.extend([height_assert, width_assert])

  # Create a random bounding box.
  #
  # Use tf.random_uniform and not numpy.random.rand as doing the former would
  # generate random numbers at graph eval time, unlike the latter which
  # generates random numbers at graph definition time.
  max_offset_height = control_flow_ops.with_dependencies(
      asserts, tf.reshape(image_height - crop_height + 1, []))
  max_offset_width = control_flow_ops.with_dependencies(
      asserts, tf.reshape(image_width - crop_width + 1, []))
  offset_height = tf.random_uniform([], maxval=max_offset_height, dtype=tf.int32)
  offset_width = tf.random_uniform([], maxval=max_offset_width, dtype=tf.int32)

  cropped_image = _crop(image, offset_height, offset_width, crop_height, crop_width)
  cropped_label = _crop(label, offset_height, offset_width, crop_height, crop_width)
  return cropped_image, cropped_label


def tf_random_rotate(image, label, max_angle=90.0):
  random_angle = (tf.to_float(tf.random_uniform([1]))[0] * 2 - 1) * max_angle / 180.0 * 3.14
  image = tf.expand_dims(image, 0)
  label = tf.expand_dims(label, 0)
  image = tf.contrib.image.rotate(image, random_angle)
  label = tf.contrib.image.rotate(label, random_angle)
  image = tf.squeeze(image, 0)
  label = tf.squeeze(label, 0)
  return image, label


def tf_random_aspect_resize(image, label, low_val=1.0, upper_val=1.5):
  shape = tf.shape(image)
  height = shape[0]
  width = shape[1]

  # 1~1.5
  which_side = tf.to_float(tf.random_uniform([1]))[0]
  multi_val = tf.to_float(tf.random_uniform([1]))[0] * (upper_val - low_val) + low_val

  new_height = tf.cond(which_side > 0.5, lambda: tf.to_float(height), lambda: tf.to_float(height) * multi_val)
  new_width = tf.cond(which_side <= 0.5, lambda: tf.to_float(width), lambda: tf.to_float(width) * multi_val)

  new_height = tf.to_int32(new_height)
  new_width = tf.to_int32(new_width)
  
  image = tf.expand_dims(image, 0)
  label = tf.expand_dims(label, 0)
  resized_image = tf.image.resize_bilinear(image, [new_height, new_width], align_corners=False)
  resized_image = tf.cast(resized_image, tf.uint8)
  resized_label = tf.image.resize_nearest_neighbor(label, [new_height, new_width], align_corners=False)
  resized_label = tf.cast(resized_label, tf.uint8)
  resized_image = tf.squeeze(resized_image, 0)
  resized_label = tf.squeeze(resized_label, 0)
  return resized_image, resized_label


def _smallest_size_at_least(height, width, smallest_side):
  smallest_side = tf.convert_to_tensor(smallest_side, dtype=tf.int32)
  
  height = tf.to_float(height)
  width = tf.to_float(width)
  smallest_side = tf.to_float(smallest_side)
  
  scale = tf.cond(tf.greater(height, width), lambda: smallest_side / width, lambda: smallest_side / height)
  new_height = tf.to_int32(height * scale)
  new_width = tf.to_int32(width * scale)
  return new_height, new_width


def tf_aspect_preserving_resize(image, label, smallest_side):
  smallest_side = tf.convert_to_tensor(smallest_side, dtype=tf.int32)
  
  shape = tf.shape(image)
  height = shape[0]
  width = shape[1]
  new_height, new_width = _smallest_size_at_least(height, width, smallest_side)
  
  new_height = tf.maximum(new_height, smallest_side)
  new_width = tf.maximum(new_width, smallest_side)
  
  image = tf.expand_dims(image, 0)
  label = tf.expand_dims(label, 0)
  resized_image = tf.image.resize_bilinear(image, [new_height, new_width], align_corners=False)
  resized_image = tf.cast(resized_image, tf.uint8)
  resized_image = tf.squeeze(resized_image, 0)
  
  resized_label = tf.image.resize_nearest_neighbor(label, [new_height, new_width], align_corners=False)
  resized_label = tf.cast(resized_label, tf.uint8)
  resized_label = tf.squeeze(resized_label, 0)
  return resized_image, resized_label


def tf_mirror(image, label):
  val_lr = tf.to_float(tf.random_uniform([1]))[0]
  image = tf.cond(val_lr > 0.5, lambda: tf.image.flip_left_right(image), lambda: image)
  label = tf.cond(val_lr > 0.5, lambda: tf.image.flip_left_right(label), lambda: label)
  
  return image, label
